# Standard imports
import sys
import json
import os
import pickle
import uuid
from ast import literal_eval
from functools import lru_cache
from typing import Any, Iterable

# Qt imports
from PySide2.QtWidgets import (
    QWidget,
    QTreeView,
    QFrame,
    QToolButton,
    QPushButton,
    QCompleter,
    QDialog,
    QFileDialog,
    QApplication,
    QStyledItemDelegate,
    QToolBar,
    QAbstractItemView,
    QHeaderView,
    QComboBox,
    QSizePolicy,
    QMessageBox,
    QHBoxLayout,
    QVBoxLayout,
    QMenu,
    QStyle,
)
from PySide2.QtCore import (
    Qt,
    QObject,
    Signal,
    Slot,
    QDir,
    QAbstractItemModel,
    QModelIndex,
    QMimeData,
    QEvent,
    QStandardPaths,
    QSize,
    QFileInfo,
    QSettings,
    QRect,
)
from PySide2.QtGui import QPainter, QPalette, QFont, QPen, QBrush

# Custom imports
from cutevariant.gui import style, plugin, FIcon
from cutevariant.core import sql, get_sql_connection
from cutevariant.core.vql import parse_one_vql
from cutevariant.core.querybuilder import (
    build_vql_query,
    fields_to_vql,
    wordset_data_to_vql,
    WORDSET_FUNC_NAME,
)
import cutevariant.commons as cm

LOGGER = cm.logger()


@lru_cache()
def prepare_fields(conn):
    """Prepares a list of columns on which filters can be applied"""
    columns = []
    samples = [sample["name"] for sample in sql.get_samples(conn)]

    for field in sql.get_fields(conn):
        if field["category"] == "samples":
            # Replace the name for samples with:
            # ("sample", <individual_id>, <field_name>)
            # Ex: with a "name" equal to "ps":
            # {'id': 48, 'name': ('sample', 'HG001', 'ps'), 'category': 'samples',
            # 'type': 'str', 'description': ''}
            for sample in samples:
                temp_field = field.copy()
                temp_field["name"] = ("sample", sample, field["name"])
                columns.append(temp_field)
        else:
            columns.append(field)
    return columns


class BaseField(QFrame):
    """Base class for all editor widgets.

    Editor widgets are used in FilterDelegate to display different kind of
    editors according to field type.

    Inherit from this class if you want a custom field editor by overriding
    `set_value` and `get_value`.

    See Also:
         :meth:`FilterDelegate`
    """

    def __init__(self, parent=None):
        super().__init__(parent)
        # style hack : Set background as same as selection in the view
        self.setAutoFillBackground(True)
        self.setBackgroundRole(QPalette.Highlight)

    def set_value(self, value):
        raise NotImplementedError

    def get_value(self):
        raise NotImplementedError

    def set_widget(self, widget):
        """Setup a layout with a widget

        Typically, it is used to add user input widget to the item
        (QSpinBox, QComboBox, etc.)

        Args:
            widget (QWidget)
        """
        self.widget = widget
        h_layout = QHBoxLayout()
        h_layout.addWidget(widget)
        h_layout.setContentsMargins(0, 0, 0, 0)
        h_layout.setSpacing(0)
        self.setLayout(h_layout)


class IntegerField(BaseField):
    """Editor for integer value

    Attributes:
        spin_box (QSpinBox)
    """

    def __init__(self, parent=None):
        super().__init__(parent)
        self.spin_box = QSpinBox()
        self.set_widget(self.spin_box)
        self.spin_box.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)

    def set_value(self, value: int):
        self.spin_box.setValue(value)

    def get_value(self) -> int:
        return self.spin_box.value()

    def set_range(self, min_, max_):
        """ Limit editor with a range of value """
        self.spin_box.setRange(min_, max_)


class FloatField(BaseField):
    """Editor for floating point value

    Attributes:
        spin_box (QDoubleSpinBox)
    """

    def __init__(self, parent=None):
        super().__init__(parent)
        self.spin_box = QDoubleSpinBox()
        self.set_widget(self.spin_box)
        self.spin_box.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)

    def set_value(self, value: int):
        self.spin_box.setValue(value)

    def get_value(self) -> float:
        return self.spin_box.value()

    def set_range(self, min_, max_):
        self.spin_box.setRange(min_, max_)


class StrField(BaseField):
    """Editor for string value

    Attributes:
        edit (QLineEdit)
    """

    def __init__(self, parent=None):
        super().__init__(parent)
        self.edit = QLineEdit()
        self.edit.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        self.set_widget(self.edit)

    def set_value(self, value: Any):
        """Set displayed value in the lineEdit of the editor

        Notes: TODO: TO be updated/removed if WORDSET is a real operator
            For the following query:
            SELECT favorite,chr,pos,ref,alt FROM variants where gene IN WORDSET['aa']

            We need to show to the user only "WORDSET['aa']"; not the internal
            representation: ("WORDSET", "aa"). This must be done in
            the model that return one representation or the other according to
            DisplayRole or UserRole.

            Here we get the internal representation (tuple) from the UserRole
            of the model.
            So we need to cast it via wordset_data_to_vql().

            The opposite is made in get_value() that MUST return internal
            representation.
        """
        # TODO: ugly cast to handle tuple corresponding to (WORDSET, name)
        if isinstance(value, tuple) and value[0] == WORDSET_FUNC_NAME:
            value = wordset_data_to_vql(value)

        self.edit.setText(str(value))

    def get_value(self) -> Any:
        """Return string or float/int for numeric values"""
        value = self.edit.text()

        # TODO: ugly cast to return tuple corresponding to (WORDSET, name)
        # Please just leave pass in this section if WORDSET is converted to normal operator
        import re

        match = re.search(r"WORDSET\[['\"](.*)['\"]\]", value)
        if match:
            return "WORDSET", match[1]

        if value.isdigit():
            return int(value)

        if value.isdecimal():
            return float(value)
        return value

    def set_completer(self, completer: QCompleter):
        """Set a completer to autocomplete value"""
        self.edit.setCompleter(completer)


class IterableField(StrField):
    """Editor for iterables in string form.

    Attributes:
        edit (QLineEdit)
    """

    def set_value(self, value: Iterable):
        """Set displayed value in the lineEdit of the editor"""
        print("Iterable field set value ?", value, type(value))
        super().set_value(value)

    def get_value(self) -> tuple:
        """Return the value of the field in tuple type

        Notes:
            Try to cast the follwing terms into tuples:
            - 'a, b' => ('a', 'b')
            - 'a,b' => ...
            - '(a, b' => ...
            - 'a,b)" => ...
            - ('a', 'b') => ...
            - (1, 2) => (1, 2)

        Returns:
            (tuple): Casted from string
        """
        value = self.edit.text()

        try:
            if "," in value and ("(" not in value or ")" not in value):
                value = value.replace("(", "").replace(")", "").replace(", ", ",")
                value = tuple(value.split(","))
                # print("splitted value", value)
                return value

            # Cast proper tuple
            return literal_eval(value)
        except ValueError:
            pass
        # Cast to str
        return super().get_value()


class ComboField(BaseField):
    """Editor for string value

    Notes:
        Items are added in FilterDelegate.createEditor()

    Attributes:
        edit (QLineEdit)
    """

    def __init__(self, parent=None):
        super().__init__(parent)
        self.edit = QComboBox()
        self.edit.setEditable(True)
        self.edit.setSizeAdjustPolicy(QComboBox.AdjustToContents)
        self.set_widget(self.edit)

    def set_value(self, value: str):
        """Set text of lineEdit in ComboBox"""
        # items = [self.edit.itemText(i) for i in range(self.edit.count())]
        # Set text of lineEdit via the index of the required text
        # Here we use an editable combobox with a lineEdit.
        # => Use setCurrentIndex instead of setCurrentText.
        # The last one doesn't refresh the index and thus the currentData.
        # In this case, get_value will return the currentData of the first item
        # in ComboBox regardless the item in the lineEdit.
        index = self.edit.findText(value)
        if index != -1:
            self.edit.setCurrentIndex(index)
            # /!\ Don't do this only: self.edit.setCurrentText(value) (see above)

    def get_value(self):
        """Return UserRole string"""
        # Return UserRole
        return self.edit.currentData()

    def addItems(self, words: list):
        """Set a completer to autocomplete value"""
        self.edit.clear()
        self.edit.addItems(words)


class BoolField(BaseField):
    """Editor for Boolean value

    Attributes:
        box (QCheckBox)
    """

    def __init__(self, parent=None):
        super().__init__(parent)
        self.box = QComboBox()
        # DisplayRole, UserRole
        self.box.addItem("False", False)
        self.box.addItem("True", True)
        self.set_widget(self.box)
        self.box.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)

    def set_value(self, value: bool):
        self.box.setCurrentIndex(int(value))

    def get_value(self) -> bool:
        # Return UserRole
        return self.box.currentData()


class OperatorField(BaseField):
    """Editor for Logic Value (less, greater, more than etc ...)

    Attributes:
        combo_box (QComboBox): Combobox to allow a suer to select operators.
    """

    # Symbols used in VQL vs text descriptions
    SYMBOLS = {
        "<": "less",
        "<=": "less or equal",
        ">": "greater",
        ">=": "greater or equal",
        "=": "equal",
        "!=": "not equal",
        "LIKE": "like",
        "NOT LIKE": "not like",
        "~": "regex",
        "IN": "in",
        "NOT IN": "not in",
        "IS": "is",
        "IS NOT": "is not",
    }

    def __init__(self, parent=None):
        super().__init__(parent)
        self.combo_box = QComboBox()
        self.set_widget(self.combo_box)
        self.combo_box.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)

        self._fill()

    def set_value(self, value: str):
        assert value.upper() in self.SYMBOLS
        self.combo_box.setCurrentText(self.SYMBOLS[value.upper()])

    def get_value(self) -> str:
        # Return UserRole
        return self.combo_box.currentData()

    def _fill(self):
        """Init QComboBox with all supported operators"""
        self.combo_box.clear()
        for symbol, text in self.SYMBOLS.items():
            # DisplayRole, UserRole
            self.combo_box.addItem(text, symbol)


class LogicField(BaseField):
    """Editor for logic field (And/Or)"""

    def __init__(self, parent=None):
        super().__init__(parent)
        self.box = QComboBox()

        # DisplayRole, UserRole
        self.box.addItem("AND", "AND")
        self.box.addItem("OR", "OR")

        self.box.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        self.set_widget(self.box)

    def set_value(self, value: str):

        if value.upper() == "OR":
            self.box.setCurrentIndex(1)

        else:  # AND
            self.box.setCurrentIndex(0)

    def get_value(self) -> str:
        # Return UserRole
        return self.box.currentData()


class FieldFactory(QObject):
    """FieldFactory is a factory to build BaseEditor according sql Field data

    Attributes:
        conn (sqlite3.connection)

    TODO: used only in FieldDialog => not used anymore
    """

    def __init__(self, conn):
        super().__init__()
        self.conn = conn

        self.field_types_mapping = {
            field["name"]: field["type"] for field in prepare_fields(conn)
        }

    def create(self, sql_field, parent=None):
        """Get FieldWidget according to type key of the given sql_field"""
        # sample fields are stored as tuples
        # if type(sql_field) is tuple:
        # sample = sql_field[1]
        # sql_field = sql_field[2]
        # else:
        # # Passing sample has no effect for non-sample fields
        # sample = None

        # Get field data by its name in DB
        # Don't work anymore because some fields (samples related)
        # have a tuple structure.
        # See prepare_fields()
        # field_type = sql.get_field_by_name(self.conn, sql_field)

        field_type = self.field_types_mapping.get(sql_field)
        assert field_type

        if field_type == "int":
            w = IntegerField(parent)
            # w.set_range(*sql.get_field_range(self.conn, sql_field, sample))
            return w

        if field_type == "float":
            w = FloatField(parent)
            # w.set_range(*sql.get_field_range(self.conn, sql_field, sample))
            return w

        if field_type == "str":
            return StrField(parent)

        if field_type == "bool":
            return BoolField(parent)

        return StrField(parent)


class FilterItem:
    """FilterItem is a recursive class which represent item for a FilterModel

    A tree of FilterItems can be stored by adding FilterItems recursively as children.
    Each FilterItem has a parent and a list of children.
    see https://doc.qt.io/qt-5/qtwidgets-itemviews-simpletreemodel-example.html

    Attributes:
        parent(FilterItem): item's parent
        children(list[FilterItem]): list of children
        data(any): str (logicType) or tuple/list (ConditionType).
        uuid(str):
        checked(boolean):
        type(FilterItem.LOGIC_TYPE/FilterItem.CONDITION_TYPE): Type of filter item.

    Examples:
        root = FilterItem() # Create rootItem
        root.append(FilterItem()) # Append 2 children
        root.append(FilterItem())
        root[0].append(FilterItem()) # Append 1 child to the first children
    """

    LOGIC_TYPE = 0  # Logic type is AND/OR
    CONDITION_TYPE = 1  # Condition type is (field, operator, value)

    def __init__(self, data=None, parent=None):
        """FilterItem constructor with parent as FilterItem parent

        Args:
            data(any): str (logicType) or tuple/list (ConditionType).
            parent (FilterItem): item's parent
        """
        # Item Type handling
        is_tuple = isinstance(data, (tuple, list))
        assert is_tuple or isinstance(data, str)
        self.data = list(data) if is_tuple else data
        self.type = self.CONDITION_TYPE if is_tuple else self.LOGIC_TYPE
        # Misc
        self.parent = parent
        self.children = []
        self.uuid = str(uuid.uuid1())
        self.checked = True

    def __del__(self):
        """Clear children (list[FilterItem])"""
        self.children.clear()

    def __repr__(self):
        return f"Filter Item {self.data}"

    def __getitem__(self, row):
        """Return FilterItem at the given index

        Args:
            row (int): child position

        Returns:
            FilterItem
        """
        return self.children[row]

    def append(self, item):
        """Append FilterItem child

        Args:
            item (FilterItem)
        """
        item.parent = self
        self.children.append(item)

    def insert(self, row: int, item):
        """Insert FilterItem child at a specific location

        Args:
            row (int): child index
            item (FilterItem)
        """
        item.parent = self
        self.children.insert(row, item)

    def remove(self, row: int):
        """Remove FilterItem child from a specific position

        Args:
            row (int): child index
        """
        del self.children[row]

    def row(self) -> int:
        """Return item location from his parent.

        If the item has no parent, it returns 0

        Returns:
            int: item index
        """
        if self.parent is not None:
            return self.parent.children.index(self)
        return 0

    def setRecursiveChecked(self, checked=True):
        self.checked = checked
        for child in self.children:
            child.set_recursive_check_state(checked)

    def get_field(self):
        if self.type == self.CONDITION_TYPE:
            return self.data[0]

    def get_operator(self):
        if self.type == self.CONDITION_TYPE:
            return self.data[1].upper()

    def get_value(self):
        """Get value of condition or operator value

        Returns:
            - If item is a LOGIC_FIELD, return the operator AND/OR.
            - If item is a CONDITION_TYPE, return the value of the condition (last field).

        Examples:
            For a CONDITION_TYPE FilterItem: `("chr", "IN", (10, 11))`,
            this function will return `(10, 11)`.
        """
        if self.type == self.CONDITION_TYPE:
            return self.data[2]

        if self.type == self.LOGIC_TYPE:
            return self.data

    # def get_data(self, column=0):
    #     """ get data according columns.

    #     if item is a LOGIC_FIELD, it return self.data not matter the column.
    #     If item is a CONDITION_TYPE, you can select value from tuple according columns.

    #     column 0: Field name
    #     column 1: Field operator
    #     column 2 : Field value
    #     Args:
    #         column (int)

    #     Returns:
    #         (any): Data
    #     """

    #     if column == 0:
    #         return self.checked

    #     # if column == 1 or column == 2 or column == 3:
    #     #     if self.type == self.LOGIC_TYPE:
    #     #         return self.data

    #     #     if self.type == self.CONDITION_TYPE:
    #     #         return self.data[column - 1]

    def set_field(self, value):
        """Set field part of CONDITION_TYPE item"""
        if self.type == self.CONDITION_TYPE:
            self.data[0] = value

    def set_operator(self, value):
        """Set operator part of CONDITION_TYPE item"""
        if self.type == self.CONDITION_TYPE:
            self.data[1] = value.upper()

    def set_value(self, value):
        """Set value part of CONDITION_TYPE item or value of LOGIC_TYPE item

        Called when a user validates the editor.
        """
        if self.type == self.CONDITION_TYPE:
            self.data[2] = value
            return

        # LOGIC_TYPE:
        self.data = value


class FilterModel(QAbstractItemModel):
    """Model to display filter

    The model store Query filter as a nested tree of FilterItem.
    You can access data from self.item(), edit model using self.set_data()
    and helper methods like: add_logic_item, add_condition_item and remove_item.

    Attributes:
        conn (sqlite3.connection): sqlite3 connection
        root_item (FilterItem): RootItem (invisible) to store recursive item.

    Additional roles:
        TypeRole: Items types (LOGIC_TYPE or CONDITION_TYPE)
        UniqueIdRole: Uuid of items.

    Signals:
        filtersChanged: Emitted when model data (filters) is changed.

    Examples:
        data = {"AND": [
        {"field": "ref", "operator": "=", "value": "A"},
        {
            "OR": [
                {"field": "chr", "operator": "=", "value": "chr5"},
                {"field": "chr", "operator": "=", "value": "chr3"},
            ]
        },}}
        model = FilterModel(conn)
        model.load(data)
        view = QTreeView()
        view.setModel(model)

        # Access item
        item  = model.item(view.currentIndex())
        # Add new item
        model.add_logic_item(parent = view.currentIndex())
        # Remove item
        model.remove_item(view.currentIndex())
    """

    # See self.headerData()
    _HEADERS = ["c", "field", "operator", "value", "op"]
    _MIMEDATA = "application/x-qabstractitemmodeldatalist"

    # Custom type to get FilterItem.type. See self.data()
    TypeRole = Qt.UserRole + 1
    UniqueIdRole = Qt.UserRole + 2

    filtersChanged = Signal()

    def __init__(self, conn=None, parent=None):
        super().__init__(parent)
        self.root_item = FilterItem("AND")
        self.conn = conn
        self.clear()

    @property
    def filters(self):
        return self.to_dict()

    @filters.setter
    def filters(self, filters):
        self.load(filters)

    def __del__(self):
        """Model destructor."""
        del self.root_item

    def data(self, index: QModelIndex, role=Qt.EditRole):
        """Overrided Qt methods : Return model's data according index and role

        Warning:
            FilterDelegate.createEditor and setEditorData must use UserRole!
            The displayed elements are displayed from FilterItem with DisplayRole!
            Field* take ONLY UserRoles and convert them into something that can be
            showed to a user.

        Args:
            index (QModelIndex): index of item
            role (Qt.Role)

        Returns:
            Any type: Return value
        """
        if not index.isValid():
            return

        item = self.item(index)

        if role in (Qt.DisplayRole, Qt.EditRole, Qt.UserRole):
            if index.column() == 0:
                return item.checked if role == Qt.UserRole else str(item.checked)

            if index.column() == 1:
                if item.type == FilterItem.CONDITION_TYPE:
                    return fields_to_vql(item.get_field())

                if item.type == FilterItem.LOGIC_TYPE:
                    val = item.get_value()
                    return val if role == Qt.UserRole else str(val)

            if item.type != FilterItem.CONDITION_TYPE:
                return

            if index.column() == 2:
                operator = item.get_operator()
                return operator if role == Qt.UserRole else str(operator)

            if index.column() == 3:
                val = item.get_value()
                # TODO: WORDSET handling here. To be modified if WORDSET is a real operator
                # Same way that for fields formatting in VQL form for column 1
                if (
                    role in (Qt.EditRole, Qt.DisplayRole)
                    and isinstance(val, tuple)
                    and val[0] == WORDSET_FUNC_NAME
                ):
                    return wordset_data_to_vql(val)

                return val if role == Qt.UserRole else str(val)

        if role == FilterModel.TypeRole:
            # Return item type
            return item.type

        if role == FilterModel.UniqueIdRole:
            return item.uuid

        return

        # if role == Qt.DisplayRole and index.column() == 1:
        #     data = self.item(index).get_data(index.column())
        #     return str(data)

        # if role in (Qt.DecorationRole, Qt.DisplayRole) and index.column() == 2:
        #     # Special case to display an icon instead of a number for gt fields
        #     field = self.item(index).get_data(0)
        #     if type(field) is tuple and field[2] == "gt":
        #         if role == Qt.DecorationRole:
        #             gt = self.item(index).get_data(index.column())
        #             return QIcon(GenotypeField.GENOTYPES[gt])
        #         else:
        #             # don't display any text
        #             return None
        # if role == Qt.DisplayRole or role == Qt.EditRole:
        #     #  Display data
        #     item = self.item(index)
        #     return item.get_data(index.column())

        # if role == Qt.TextAlignmentRole:
        #     #  Adjust text alignement
        #     if index.column() == 0:
        #         return int(Qt.AlignVCenter) + int(Qt.AlignLeft)
        #     if index.column() == 1:
        #         return Qt.AlignCenter
        #     if index.column() == 2:
        #         return int(Qt.AlignVCenter) + int(Qt.AlignLeft)

        # if role == Qt.FontRole:
        #     #  Make LogicItem as bold
        #     if self.item(index).type == FilterItem.LOGIC_TYPE:
        #         font = QFont()
        #         font.setBold(True)
        #         return font

    def setData(self, index, value, role=Qt.UserRole):
        """Overrided Qt methods: Set value of FilterItem present at the given index.

        This method is called from FilterDelegate when edition has been done.

        Warning:
            FilterDelegate.createEditor and setEditorData must use UserRole!
            The displayed elements are displayed from FilterItem with DisplayRole!
            Field* take ONLY UserRoles and convert them into something that can be
            showed to a user.

        Args:
            index (QModelIndex)
            value (any): new value
            role (Qt.ItemDataRole): Qt.UserRole or Qt.CheckStateRole

        Returns:
            bool: Return True if success otherwise return False
        """
        if not index.isValid():
            return False

        if role in (Qt.DisplayRole, Qt.EditRole, Qt.UserRole):
            item = self.item(index)

            if index.column() == 0:
                item.checked = bool(value)

            if index.column() == 1:
                if item.type == FilterItem.LOGIC_TYPE:
                    item.set_value(value)

                if item.type == FilterItem.CONDITION_TYPE:
                    item.set_field(value)

            if item.type == FilterItem.CONDITION_TYPE:

                if index.column() == 2:
                    item.set_operator(value)

                if index.column() == 3:
                    item.set_value(value)

            self.filtersChanged.emit()
            # just one item is changed
            self.dataChanged.emit(index, index, role)
            return True

        if role == Qt.CheckStateRole:
            self.set_recursive_check_state(index, bool(value))
            self.filtersChanged.emit()
            # just one item is changed
            self.dataChanged.emit(index, index, role)
            return True

        return False

    def headerData(self, section, orientation, role=Qt.DisplayRole):
        """Return header data

        Args:
            section (integer): row
            orientation (Qt.Orientation): Vertical or horizontal header
            role (Qt.ItemDataRole, optional): data role

        Returns:
            Any type of data
        """
        if role == Qt.DisplayRole:
            if orientation == Qt.Horizontal:
                return self._HEADERS[section]

    def is_last(self, index: QModelIndex()) -> bool:
        """Return True if index is the last in the row
        This is used by draw_branch
        """
        if index == QModelIndex():
            return False

        return index.row() == index.model().rowCount(index.parent()) - 1

    def index(self, row, column, parent=QModelIndex()) -> QModelIndex:
        """Overrided Qt methods: create index according row, column and parent

        Usefull for dataChanged signal

        Returns:
            QModelIndex
        """
        if not self.hasIndex(row, column, parent):
            return QModelIndex()

        if not parent.isValid():  # If no parent, then parent is the root item
            parent_item = self.root_item

        else:
            parent_item = parent.internalPointer()

        child_item = parent_item[row]
        if child_item:
            return self.createIndex(row, column, child_item)
        else:
            return QModelIndex()

    def parent(self, index: QModelIndex) -> QModelIndex:
        """Overrided Qt methods: Create parent from index """
        if not index.isValid():
            return QModelIndex()

        child_item = index.internalPointer()

        parent_item = child_item.parent

        if parent_item == self.root_item:
            return QModelIndex()

        return self.createIndex(parent_item.row(), 0, parent_item)

    def clear(self):
        """Clear Model"""
        self.beginResetModel()
        self.root_item.children.clear()
        # Load first default item
        self.root_item.append(FilterItem("AND"))

        self.endResetModel()

    def load(self, data: dict):
        """load model from dict

        dict should be a nested dictionnary of condition. For example:
        data = {"AND": [
        {"field": "ref", "operator": "=", "value": "A"},
        {
            "OR": [
                {"field": "chr", "operator": "=", "value": "chr5"},
                {"field": "chr", "operator": "=", "value": "chr3"},
            ]
        },}}
        Args:
            data (TYPE): Description
        """
        self.beginResetModel()
        if data:
            self.root_item.children.clear()
            self.root_item.append(self.to_item(data))
        self.endResetModel()

    def to_item(self, data: dict) -> FilterItem:
        """Recursive function to build a nested FilterItem structure from dict data"""
        if len(data) == 1:  # logic item
            operator = list(data.keys())[0]
            item = FilterItem(operator)
            [item.append(self.to_item(k)) for k in data[operator]]
        else:  # condition item
            item = FilterItem((data["field"], data["operator"], data["value"]))
        return item

    def to_dict(self, item=None) -> dict:
        """Recursive function to build a nested dictionnary from FilterItem structure

        Notes:
            We use data from FilterItems; i.e. the equivalent of UserRole data.
        """

        if len(self.root_item.children) == 0:
            return {}

        if item is None:
            item = self.root_item[0]

        if item.type == FilterItem.LOGIC_TYPE and item.checked is True:
            # Return dict with operator as key and item as value
            operator_data = [
                self.to_dict(child) for child in item.children if child.checked is True
            ]
            return {item.get_value(): operator_data}

        if item.type == FilterItem.CONDITION_TYPE:
            return {
                "field": item.get_field(),
                "operator": item.get_operator(),
                "value": item.get_value(),
            }

    def add_logic_item(self, value="AND", parent=QModelIndex()):
        """Add logic item

        Args:
            value (str): Can be "AND" or "OR"
            parent (QModelIndex): parent index
        """

        #  Skip if parent is a condition type
        if self.item(parent).type == FilterItem.CONDITION_TYPE:
            return

        self.beginInsertRows(parent, 0, 0)
        self.item(parent).insert(0, FilterItem(data=value))
        self.endInsertRows()
        self.filtersChanged.emit()

    def add_condition_item(self, value=("chr", ">", "100"), parent=QModelIndex()):
        """Add condition item

        Args:
            value (tuple): Condition data (field, operator, value)
            parent (QModelIndex): Parent index
        """
        # Skip if parent is a condition type
        if self.item(parent).type == FilterItem.CONDITION_TYPE:
            return

        row = self.rowCount(parent)
        self.beginInsertRows(parent, row - 1, row - 1)
        item = FilterItem(data=value)
        self.item(parent).append(item)
        self.endInsertRows()
        self.filtersChanged.emit()

    def remove_item(self, index):
        """Remove Item

        Args:
            index (QModelIndex): item index
        """
        if index.isValid():
            self.beginRemoveRows(index.parent(), index.row(), index.row())
            self.item(index).parent.remove(index.row())
            self.endRemoveRows()
            self.filtersChanged.emit()

    def rowCount(self, parent=QModelIndex()) -> int:
        """Overrided Qt methods: return row count according parent """
        if parent.column() > 0:
            return 0

        if not parent.isValid():
            parent_item = self.root_item
        else:
            parent_item = parent.internalPointer()

        return len(parent_item.children)

    def columnCount(self, parent=QModelIndex()) -> int:
        """ Overrided Qt methods: return column count according parent """

        return 5

    def flags(self, index) -> Qt.ItemFlags:
        """ Overrided Qt methods: return Qt flags to make item editable and selectable """

        if not index.isValid():
            return 0

        item = index.internalPointer()

        if index.column() == 0 or index.column() == 4:
            return Qt.ItemIsSelectable | Qt.ItemIsEnabled

        if item.type == FilterItem.LOGIC_TYPE and index.column() != 1:
            return Qt.ItemIsSelectable | Qt.ItemIsEnabled

        if item.type == FilterItem.LOGIC_TYPE and index.column() == 1:
            return (
                Qt.ItemIsSelectable
                | Qt.ItemIsEditable
                | Qt.ItemIsEnabled
                | Qt.ItemIsDragEnabled
                | Qt.ItemIsDropEnabled
            )

        if item.type == FilterItem.CONDITION_TYPE:
            return (
                Qt.ItemIsSelectable
                | Qt.ItemIsEditable
                | Qt.ItemIsEnabled
                | Qt.ItemIsDragEnabled
            )

        return Qt.ItemIsSelectable | Qt.ItemIsEditable | Qt.ItemIsEnabled

    def item(self, index: QModelIndex) -> FilterItem:
        """Return Filter Item from model index

        Args:
            index (QModelIndex)

        Returns:
            FilterItem
        """
        if index.isValid():
            return index.internalPointer()
        else:
            return self.root_item

    def moveRow(
        self,
        sourceParent: QModelIndex,
        sourceRow: int,
        destinationParent: QModelIndex,
        destinationChild: int,
    ) -> bool:
        """Overrided Qt methods : Move an item from source to destination index
        Args:
            sourceParent (QModelIndex): parent of souce item
            sourceRow (int): index position of source item
            destinationParent (QModelIndex): parent od destination item
            destinationChild (int): index position of destination item

        Returns:
            bool: Return True if success otherwise retur False
        """
        parent_source_item = self.item(sourceParent)
        parent_destination_item = self.item(destinationParent)

        #  if destination is - 1, it's mean we should append the item at the end of children
        if destinationChild < 0:
            if sourceParent == destinationParent:
                return False
            else:
                destinationChild = len(parent_destination_item.children)

        # Don't move same same Item
        if sourceParent == destinationParent and sourceRow == destinationChild:
            return False

        self.beginMoveRows(
            sourceParent, sourceRow, sourceRow, destinationParent, destinationChild
        )
        item = parent_source_item.children.pop(sourceRow)
        parent_destination_item.insert(destinationChild, item)
        self.endMoveRows()

        self.filtersChanged.emit()
        return True

    def supportedDropActions(self) -> Qt.DropAction:
        """Overrided from Qt. Return supported drop action by the model

        Returns:
            Qt.DropAction
        """
        return Qt.MoveAction

    def dropMimeData(self, data, action, row, column, parent) -> bool:
        """Overrided Qt methods: This method is called when item is dropped by drag/drop.
        data is QMimeData and it contains a pickle serialization of current dragging item.
        Get back item by unserialize data.data().

        Args:
            data (QMimeData)
            action (Qt.DropAction)
            row (int): row destination
            column (int): column destination ( not used)
            parent (QModelIndex): parent destination

        Returns:
            bool: return True if success otherwise return False
        """
        if action != Qt.MoveAction:
            return False

        if not data.data(self._MIMEDATA):
            return False

        # Unserialize
        item = pickle.loads(data.data(self._MIMEDATA).data())

        # Get index from item
        source_parent = self.match(
            self.index(0, 0),
            FilterModel.UniqueIdRole,
            item.parent.uuid,
            1,
            Qt.MatchRecursive,
        )

        if source_parent:
            source_parent = source_parent[0]
            return self.moveRow(source_parent, item.row(), parent, row)

        return False

    def mimeData(self, indexes) -> QMimeData:
        """Serialize item from indexes into a QMimeData
        Currently, it serializes only the first index from t he list.
        Args:
            indexes (list<QModelIndex>)

        Returns:
            QMimeData
            ..see: self.dropMimeData
        """
        if not indexes:
            return

        data = QMimeData(self._MIMEDATA)
        serialization = QByteArray(pickle.dumps(self.item(indexes[0])))
        data.setData(self._MIMEDATA, serialization)
        return data

    def set_recursive_check_state(self, index, checked=True):
        """Recursive check of all subfilters"""
        if not index.isValid():
            return

        item = self.item(index)

        item.checked = checked
        start = self.index(index.row(), 0, index.parent())
        end = self.index(index.row(), self.columnCount() - 1, index.parent())

        # Update specific changed item
        self.dataChanged.emit(start, end)

        for row in range(self.rowCount(index)):
            cindex = self.index(row, 0, index)
            self.set_recursive_check_state(cindex, checked)

    def to_json(self, filename: str):

        with open(filename, "w") as file:
            json.dump(self.to_dict(), file)

    def from_json(self, filename: str):
        with open(filename, "r") as file:
            data = json.load(file)
            self.filters = data


class FilterDelegate(QStyledItemDelegate):
    """FilterDelegate is used to create widget editor for the model inside the view.

    Notes:
        Without a delegate, the view cannot display editor when user double clicks
        on a cell.

        Editors are created from self.createEditor.
        FilterModel data are read and written respectively with setEditorData and
        setModelData.

        The view has 5 columns, enumerated with the following names:

        - COLUMN_CHECKBOX = 0
        - COLUMN_FIELD = 1
        - COLUMN_LOGIC = 1
        - COLUMN_OPERATOR = 2
        - COLUMN_VALUE = 3
        - COLUMN_REMOVE = 4

    Examples:
        view = QTreeView()
        model = FilterModel()
        delegate = FilterDelegate()
        view.setModel(model)
        view.setItemDelegate(delegate)
    """

    COLUMN_CHECKBOX = 0
    COLUMN_FIELD = 1
    COLUMN_LOGIC = 1
    COLUMN_OPERATOR = 2
    COLUMN_VALUE = 3
    COLUMN_REMOVE = 4

    def __init__(self, parent=None):
        super().__init__(parent)

        self.add_icon = FIcon(0xF0704)
        self.group_icon = FIcon(0xF0704)
        self.rem_icon = FIcon(0xF0156, "red")

        self.eye_on = FIcon(0xF06D0)
        self.eye_off = FIcon(0xF06D1)

        self.icon_size = QSize(20, 20)

        self.indentation = 12
        self.branch_width = 8

    def createEditor(self, parent, option, index: QModelIndex) -> QWidget:
        """Overrided from Qt. Create an editor for the selected column.

        The editor is based on the selected column and on the type of FilterItem
        (LOGIC_TYPE or CONDITION_TYPE). It is also based on the selected SQL field,
        and on the SQL operator.

        Args:
            parent (QWidget): widget's parent
            option (QStyleOptionViewItem)
            index (QModelIndex)

        Returns:
            QWidget: a editor with set_value and get_value methods
        """
        model = index.model()
        item = model.item(index)
        # Get current sql connection
        conn = model.conn

        if index.column() == 1:
            if item.type == FilterItem.LOGIC_TYPE:
                # AND/OR logic operators
                return LogicField(parent)

            if item.type == FilterItem.CONDITION_TYPE:
                # Display all fields of the database
                widget = ComboField(parent)
                # LOGGER.debug(prepare_fields.cache_info())
                # Add items
                for field in prepare_fields(conn):
                    # Cast ('sample', 'HG001', 'gt') to sample['HG001'].gt
                    # Leave "chr" to "chr"
                    text = fields_to_vql(field["name"])
                    widget.edit.addItem(text, field["name"])

                return widget

        if index.column() == 2:
            return OperatorField(parent)

        if index.column() == 3:
            if model.item(index).get_operator() in ("IN", "NOT IN"):
                # Tuple value is expecitem required
                return IterableField(parent)
            # Basic string or int or float
            return StrField(parent)
            # Dynamic field according to database type
            # return FieldFactory(conn).create(model.item(index).get_field(), parent=parent)

        return super().createEditor(parent, option, index)

    def setEditorData(self, editor: QWidget, index: QModelIndex):
        """Overrided from Qt: Read data from model (FilterItem) and set editor data

        TL;DR: populate the editor with the data from the model.

        Sets the contents of the given editor, with the data of the item at the
        given index.

        Currently, it calls BaseEditor.set_value() methods

        See Also:
            :meth:`setModelData` for the opposite function (set model data)

        Args:
            editor (QWidget)
            index (QModelindex)
        """
        # print("SET editor val from model:")
        # roles = (Qt.UserRole, Qt.EditRole, Qt.DisplayRole)
        # print("user, edit, display")
        # print(";".join(str(index.data(role)) for role in roles))
        # print(";".join(str(type(index.data(role))) for role in roles))
        # print("VAL:", index.data(role=Qt.UserRole))
        # print("editor type", type(editor))

        # Set editor data from the model (from the selected FilterItem)
        # Editors expect typed values, so don't forget to use UserRole, not EditRole
        editor.set_value(index.data(role=Qt.UserRole))

    def editorEvent(self, event: QEvent, model, option, index: QModelIndex):
        """

        When editing of an item starts, this function is called with the event
        that triggered the editing, the model, the index of the item, and the
        option used for rendering the item.

        Mouse events are sent to editorEvent() even if they don't start editing
        of the item.

        This is used here to act on COLUMN_CHECKBOX and COLUMN_REMOVE

        Args:
            event:
            model:
            option:
            index:

        Returns:
            (boolean): True if event is accepted; False otherwise.

        """
        if not index.isValid():
            return False

        # Skip action with First LogicItem root item

        if event.type() == QEvent.MouseButtonPress:

            item = model.item(index)

            if index.column() == self.COLUMN_CHECKBOX and self._check_rect(
                option.rect
            ).contains(event.pos()):
                # Invert check state
                model.setData(index, not item.checked, role=Qt.CheckStateRole)
                return True

            if index.column() == self.COLUMN_REMOVE and self._check_rect(
                option.rect
            ).contains(event.pos()):
                # Remove item

                # Do not remove first elements
                if index.parent() != QModelIndex():
                    model.remove_item(index)
                return True

        # Default implementation of base method
        return False

    def setModelData(self, editor, model, index):
        """Overrided from Qt: Update the model with data from the editor.

        Currently, it calls model.setData()

        See Also:
            :meth:`setModelData` for the opposite function (set editor data)

        Args:
            editor (QWidget): editor
            model (FilterModel)
            index (QModelindex)
        """
        # val = editor.get_value()
        # print("SET data model from editor:", val, type(val))
        # Get typed data from the editor (i.e. not a string)
        # Then set this data to the FilterItem (in the corresponding attribute)
        # via its set_value() function.
        # Default: UserRole
        model.setData(index, editor.get_value())

        # super().setModelData(editor, model, index)

    # def _compute_width(self, index):

    #     if index.isValid():
    #         font = QFont()
    #         metric = QFontMetrics(font)
    #         return metric.width(str(index.data(Qt.DisplayRole)))

    #     return 50

    def sizeHint(self, option, index: QModelIndex) -> QSize:
        """Overrided from Qt. Return size of row

        Args:
            option (QStyleOptionViewItem )
            index (QModelIndex)

        Returns:
            TYPE: Description
        """

        size = QSize(option.rect.width(), 30)

        if index.column() == self.COLUMN_CHECKBOX:
            return QSize(20, 30)

        if index.column() == self.COLUMN_OPERATOR:
            return QSize(20, 30)

        if index.column() == self.COLUMN_FIELD:
            margin = self.indentation * self._compute_level(index) + self.indentation
            size.setWidth(size.width() + margin + 10)

        return size

    def _compute_level(self, index: QModelIndex):
        level = 0
        i = index.parent()
        while i.isValid():
            i = i.parent()
            level += 1

        return level

    def _draw_branch(self, painter, option, index):

        if index.parent() == QModelIndex():
            return

        item = index.model().item(index)
        parent = index.model().item(index.parent())
        level = self._compute_level(index)
        color = option.palette.color(QPalette.WindowText)

        pen = QPen(color, 1, Qt.DotLine)
        painter.setPen(pen)
        xstart = option.rect.x() + self.indentation * (level - 1)
        xend = xstart + self.branch_width

        #  draw horizontal
        painter.drawLine(
            xstart + 2, option.rect.center().y(), xend, option.rect.center().y()
        )

        #  Draw Vertical
        ## If last
        if index.model().is_last(index):
            yend = option.rect.center().y()
            painter.drawLine(xstart, option.rect.top(), xstart, yend)

        # If middle
        else:
            yend = option.rect.bottom()
            painter.drawLine(xstart, option.rect.top(), xstart, yend)

        if level > 1:
            x = xstart
            current = index
            while current != QModelIndex():
                if not index.model().is_last(current):
                    painter.drawLine(x, option.rect.top(), x, option.rect.bottom())
                current = current.parent()
                x -= self.indentation

    def paint(self, painter, option, index):

        # ======== Draw background
        item = index.model().item(index)
        is_selected = False

        if option.state & QStyle.State_Enabled:
            bg = (
                QPalette.Normal
                if option.state & QStyle.State_Active
                or option.state & QStyle.State_Selected
                else QPalette.Inactive
            )
        else:
            bg = QPalette.Disabled

        if option.state & QStyle.State_Selected:
            is_selected = True
            painter.fillRect(option.rect, option.palette.color(bg, QPalette.Highlight))

        margin = self.indentation * (self._compute_level(index))

        #  ========= Draw Checkbox
        if index.column() == self.COLUMN_CHECKBOX:

            check_icon = self.eye_on if item.checked else self.eye_off
            rect = QRect(0, 0, self.icon_size.width(), self.icon_size.height())
            rect.moveCenter(option.rect.center())
            rect.setX(4)
            painter.drawPixmap(rect.x(), rect.y(), check_icon.pixmap(self.icon_size))

        if index.column() > self.COLUMN_CHECKBOX:

            if index.column() == 1:
                self._draw_branch(painter, option, index)
            # pen = QPen(QColor("white"), 1, Qt.DotLine)
            # painter.setPen(pen)
            # # painter.drawRect(option.rect)
            # painter.drawLine(
            #     option.rect.left(),
            #     option.rect.center().y(),
            #     margin - 2,
            #     option.rect.center().y(),
            # )

            # painter.drawLine(
            #     option.rect.left(),
            #     option.rect.top(),
            #     option.rect.left(),
            #     option.rect.bottom(),
            # )

            font = QFont()
            align = Qt.AlignVCenter
            color = option.palette.color(
                QPalette.Normal if item.checked else QPalette.Disabled,
                QPalette.HighlightedText if is_selected else QPalette.WindowText,
            )

            if (
                item.type == FilterItem.LOGIC_TYPE
                and index.column() == self.COLUMN_FIELD
            ):
                font.setBold(True)
                # metric = QFontMetrics(font)
                # print(self._compute_level(index))
                # text_width = metric.boundingRect(index.data()).width()
                # #  Draw Add buttion
                # rect = QRect(0, 0, self.icon_size.width(), self.icon_size.height())
                # rect.moveCenter(
                #     QPoint(
                #         option.rect.x() + margin + text_width + 20,
                #         option.rect.center().y(),
                #     )
                # )
                # painter.drawPixmap(
                #     rect.right() - self.icon_size.width(),
                #     rect.y(),
                #     self.add_icon.pixmap(self.icon_size),
                # )
            if index.column() == self.COLUMN_FIELD:
                align |= Qt.AlignLeft

            if index.column() == self.COLUMN_OPERATOR:
                align |= Qt.AlignCenter

            if index.column() == self.COLUMN_VALUE:
                align |= Qt.AlignLeft

            painter.setFont(font)
            painter.setPen(color)
            # Indentation level

            text_rect = option.rect
            if index.column() == 1:
                xstart = option.rect.x() + margin
                text_rect.setX(xstart)

            painter.drawText(text_rect, align, index.data(Qt.DisplayRole))

            if index.column() == self.COLUMN_REMOVE and index.parent() != QModelIndex():
                rect = QRect(0, 0, self.icon_size.width(), self.icon_size.height())
                rect.moveCenter(option.rect.center())
                painter.drawPixmap(
                    rect.right() - self.icon_size.width(),
                    rect.y(),
                    self.rem_icon.pixmap(self.icon_size),
                )

        # if index.column() == 3:
        #     painter.drawPixmap(self._icon_rect(option), self.rem_icon.pixmap(self.icon_size))

        # if index.column() == 3 and item.type == FilterItem.LOGIC_TYPE:
        #     x = option.rect.right() - 20
        #     y = option.rect.center().y() - self.icon_size.height() / 2
        #     painter.drawPixmap(QRect(x,y,self.icon_size.width(), self.icon_size.height()), self.add_icon.pixmap(self.icon_size))

        # super().paint(painter, option,index)

    def _icon_rect(self, rect):
        x = rect.x()
        y = rect.center().y() - self.icon_size.height() / 2

        return QRect(x, y, self.icon_size.width(), self.icon_size.height())

    def _check_rect(self, rect):
        return QRect(rect.x(), rect.y(), rect.height(), rect.height())

    def updateEditorGeometry(self, editor, option, index):
        """Overrided from Qt. Set editor geometry

        Args:
            editor (QWidget)
            option (QStyleOptionViewItem)
            index (QModelIndex)
        """

        if index.column() == 1:
            option.rect.setLeft(
                option.rect.x() + self.indentation * (self._compute_level(index) - 1)
            )
            editor.setGeometry(option.rect)
            return

        super().updateEditorGeometry(editor, option, index)


class FieldDialog(QDialog):
    # TODO: not used anymore
    def __init__(self, conn=None, parent=None):
        super().__init__(parent)
        self.title_label = QLabel("Non title")
        self.description_label = QLabel("Description")
        self.btn_box = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel)

        self.field_box = QComboBox()
        self.field_operator = OperatorField()

        # setup combobox
        self.field_box.setEditable(True)
        # self.field_operator.setEditable(True)

        # setup label
        font = QFont()
        font.setBold(True)
        self.title_label.setFont(font)
        self.description_label.setWordWrap(True)

        v_layout = QVBoxLayout()
        v_layout.addWidget(self.title_label)
        v_layout.addWidget(self.description_label)
        v_layout.addSpacing(10)
        self.form_layout = QFormLayout()

        self.form_layout.addRow("Field", self.field_box)
        self.form_layout.addRow("Operator", self.field_operator)
        self.form_layout.addRow("Value", QSpinBox())

        v_layout.addLayout(self.form_layout)
        v_layout.addStretch(True)
        v_layout.addWidget(self.btn_box)

        self.setLayout(v_layout)

        self.setFixedSize(500, 300)

        self.field_box.currentIndexChanged.connect(self.on_field_changed)

        self.conn = conn

        self.btn_box.accepted.connect(self.accept)
        self.btn_box.rejected.connect(self.reject)

    @property
    def conn(self):
        return self._conn

    @conn.setter
    def conn(self, conn):
        self._conn = conn
        if self._conn:
            self.load_fields()

    def load_fields(self):
        """Load sql fields into combobox"""
        for field in sql.get_field_by_category(self.conn, "variants"):
            self.field_box.addItem(field["name"], field)

    def load_value_editor(self, sql_field):
        """Create a field widget according sql field name

        Args:
            sql_field (str): field name from sql field table
        """
        self.form_layout.removeRow(2)
        widget = FieldFactory(conn).create(sql_field)
        self.form_layout.addRow("value", widget)

    @Slot(int)
    def on_field_changed(self, index):
        """This method is triggered when a field has changed

        Args:
            index (int): current index from self.field_box
        """
        field = self.field_box.itemData(index)
        self.title_label.setText("{name} ({category})".format(**field))
        self.description_label.setText(field["description"])
        self.load_value_editor(field["name"])

    def get_condition(self):
        """Return current condition as a dictionnary

        Returns:
            Dictionnary exemple {"field":"chr", "operator":"=", value:5}

        """
        field = self.field_box.currentText()
        operator = self.field_operator.get_value()
        widget = self.form_layout.itemAt(5).widget()
        value = widget.get_value()

        return {"field": field, "operator": operator, "value": value}


class FiltersEditorWidget(plugin.PluginWidget):
    """Displayed widget plugin to allow creation/edition/deletion of filters"""

    ENABLE = True
    changed = Signal()

    def __init__(self, conn=None, parent=None):
        super().__init__(parent)
        self.setWindowTitle(self.tr("Filters"))

        self.settings = QSettings()
        self.view = QTreeView()
        # conn is always None here but initialized in on_open_project()

        self.model = FilterModel(conn)
        self.delegate = FilterDelegate()
        self.toolbar = QToolBar()
        self.toolbar.setIconSize(QSize(16, 16))

        # Drag & drop
        self.view.setModel(self.model)
        self.view.setItemDelegate(self.delegate)
        self.view.setDragEnabled(True)
        self.view.header().setStretchLastSection(False)
        self.view.setAcceptDrops(True)
        self.view.setExpandsOnDoubleClick(False)

        self.view.setDragDropMode(QAbstractItemView.InternalMove)
        self.view.setAlternatingRowColors(True)
        self.view.setIndentation(0)
        self.view.header().setSectionResizeMode(0, QHeaderView.ResizeToContents)
        self.view.header().setSectionResizeMode(1, QHeaderView.Stretch)
        self.view.header().setSectionResizeMode(2, QHeaderView.ResizeToContents)
        self.view.header().setSectionResizeMode(3, QHeaderView.Stretch)
        self.view.header().setSectionResizeMode(4, QHeaderView.ResizeToContents)
        self.view.setEditTriggers(QAbstractItemView.DoubleClicked)
        # Item selected in view
        self.view.selectionModel().selectionChanged.connect(self.on_selection_changed)
        self.view.header().hide()

        self.combo = QComboBox()
        self.combo.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Preferred)
        self.combo.addItem(self.tr("Current not saved filter..."))

        # self.save_button = QToolButton()
        # self.save_button.setIcon(FIcon(0xF0193))
        # self.save_button.setToolTip(self.tr("Save the current filter"))
        # self.save_button.clicked.connect(self.on_save_filters)
        # self.del_button = QToolButton()
        # self.del_button.setDefaultAction(
        #     QAction(FIcon(0xF0A7A), self.tr("Delete the filter"))
        # )
        # self.del_button.clicked.connect(self.on_delete_item)
        # Adjust heights
        # self.combo.setMinimumHeight(30)
        # self.save_button.setMinimumHeight(30)
        # self.del_button.setMinimumHeight(30)

        self.add_filter_button = QPushButton("Add Filter")
        self.add_filter_button.setFlat(True)
        self.add_group_button = QPushButton("Add Group")
        self.add_group_button.setFlat(True)
        self.add_filter_button.clicked.connect(self.on_add_condition)
        self.add_group_button.clicked.connect(self.on_add_logic)

        blayout = QHBoxLayout()
        blayout.addWidget(self.add_filter_button)
        blayout.addWidget(self.add_group_button)

        self.apply_button = QPushButton(FIcon(0xF0233), self.tr("Apply filter"))
        self.apply_button.clicked.connect(self.on_filters_changed)

        # setup Menu
        self.toolbar.addWidget(self.combo)

        self.file_menu = QMenu()
        self.file_menu.addAction(FIcon(0xF0818), "Save ", self.on_save)
        self.file_menu.addAction(FIcon(0xF0CFC), "Save as ...", self.on_save_as)
        self.file_menu.addAction(FIcon(0xF0156), "Clear", self.model.clear)
        self.file_menu.addAction(FIcon(0xF0A7A, "red"), "Delete ...", self.on_delete)
        self.file_menu.addSeparator()
        self.file_menu.addAction(
            FIcon(0xF0DCE), "set default directory ...", self.on_set_directory
        )

        self.menu_button = QToolButton()
        self.menu_button.setIcon(FIcon(0xF035C))
        self.menu_button.setMenu(self.file_menu)
        self.menu_button.setPopupMode(QToolButton.InstantPopup)
        self.toolbar.addWidget(self.menu_button)

        layout = QVBoxLayout()

        layout.addWidget(self.toolbar)
        layout.addWidget(self.view)
        layout.addLayout(blayout)
        layout.addWidget(self.apply_button)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(1)
        self.setLayout(layout)

        self.combo.currentIndexChanged.connect(self.on_combo_changed)

        # self.model.filtersChanged.connect(self.on_filters_changed)

    @property
    def filters(self):
        return self.model.filters

    @filters.setter
    def filters(self, filters):
        self.model.filters = filters

    def on_open_project(self, conn):
        """Overrided from PluginWidget"""
        self.model.conn = conn
        self.conn = conn

        # Clear lru_cache
        prepare_fields.cache_clear()
        self.load_combo()
        self.on_combo_changed(0)
        self.on_refresh()

    # def load_predefined_filters(self):
    #     """Load user and software defined filters)

    #     - Software defined filters are loaded from "filter.json" embedded
    #         in the current plugin folder.
    #     - User defined filters are loaded from app settings, under the key
    #         `plugins/filters_editor/filters`.
    #     """
    #     json_file_path = os.path.join(os.path.dirname(__file__), "filters.json")

    #     filters = dict()
    #     # Open from embedded JSON and load software filters
    #     if os.path.isfile(json_file_path):
    #         with open(json_file_path, encoding="utf8") as f_d:
    #             try:
    #                 filters = json.loads(f_d.read())
    #             except json.decoder.JSONDecodeError as e:
    #                 LOGGER.exception(e)
    #         LOGGER.debug("Loaded predefined filters:", filters)

    #     # Overwrite software filters with user defined filters
    #     self.settings.beginGroup("plugins/filters_editor/filters")
    #     filters.update(
    #         {
    #             filter_name: self.settings.value(filter_name)
    #             for filter_name in self.settings.childKeys()
    #         }
    #     )
    #     self.settings.endGroup()

    #     # Load in combobox
    #     for filter_name, filter_query in filters.items():
    #         vql_obj = parse_one_vql(filter_query)
    #         self.combo.addItem(filter_name, vql_obj["filters"])

    def on_refresh(self):
        """Overrided"""
        if self.filters == self.mainwindow.state.filters:
            # No change in filters = no refresh
            return

        # Filters changed: Update UserRole of default filter and select it

        self.refresh_buttons()
        self._update_view_geometry()

    @property
    def filter_path(self):
        """ Return filter path from settings """
        settings = QSettings()
        settings.beginGroup(self.plugin_name)
        return settings.value(
            "filter_path", QStandardPaths.writableLocation(QStandardPaths.DataLocation)
        )

    @filter_path.setter
    def filter_path(self, value: str):
        settings = QSettings()
        settings.beginGroup(self.plugin_name)
        return settings.setValue("filter_path", value)

    def refresh_buttons(self):
        """Actualize the enable states of Add/Del buttons"""

        if self.filters:
            # Data

            # Add button: Is an item selected ?
            index = self.view.currentIndex()
            if index.isValid() and self.model.item(index).type == FilterItem.LOGIC_TYPE:
                self.add_filter_button.setEnabled(True)
                self.add_group_button.setEnabled(True)
            else:
                # item is CONDITION_TYPE or there is no item selected (because of deletion)
                self.add_filter_button.setEnabled(False)
                self.add_group_button.setEnabled(False)

    def load_combo(self):

        self.combo.clear()

        folder = QDir(self.filter_path)
        for index, file_info in enumerate(
            folder.entryInfoList(["*.filter.json"], QDir.Files)
        ):
            self.combo.addItem(
                FIcon(0xF0E29),
                str(file_info.baseName()),
                str(file_info.absoluteFilePath()),
            )

            #  Security
            if index > 1000:
                LOGGER.warning("Too many file filters. Some filters are not imported")
                break

    def on_filters_changed(self):
        """Triggered when filters changed FROM THIS plugin

        Set the filters of the mainwindow and trigger a refresh of all plugins.
        """
        if self.mainwindow and self.filters != self.mainwindow.state.filters:
            # Refresh other plugins only if the filters are modified
            self.mainwindow.state.filters = self.filters
            self.mainwindow.refresh_plugins(sender=self)

        self.refresh_buttons()

    def on_add_logic(self):
        """Add logic item to the current selected index"""
        index = self.view.currentIndex()
        if index:
            self.model.add_logic_item(parent=index)
            # self.view.setFirstColumnSpanned(0, index.parent(), True)

            self._update_view_geometry()

    def on_save_as(self):

        dialog = QFileDialog()
        dialog.setDirectory(
            QStandardPaths.writableLocation(QStandardPaths.DataLocation)
        )
        dialog.setDefaultSuffix("filter.json")
        dialog.setNameFilter("*.filter.json")
        dialog.setAcceptMode(QFileDialog.AcceptSave)

        if dialog.exec_():
            if dialog.selectedFiles():
                filename = dialog.selectedFiles()[0]
                self.model.to_json(filename)
                self.load_combo()
                file_info = QFileInfo(filename)
                self.combo.setCurrentText(file_info.baseName())

    def on_save(self):

        if self.combo.currentData() is None:
            self.on_save_as()
            return

        msgBox = QMessageBox()
        msgBox.setText("The filter has been modified.")
        msgBox.setInformativeText("Do you want to override the file ?")
        msgBox.setStandardButtons(QMessageBox.Save | QMessageBox.Discard)
        msgBox.setDefaultButton(QMessageBox.Save)

        if msgBox.exec_():
            filename = self.combo.currentData()
            if os.path.exists(filename):
                self.model.to_json(filename)

    def on_set_directory(self):

        dialog = QFileDialog()
        dialog.setDirectory(self.filter_path)
        dialog.setAcceptMode(QFileDialog.AcceptOpen)
        dialog.setFileMode(QFileDialog.DirectoryOnly)
        if dialog.exec_():
            if dialog.selectedFiles():
                filename = dialog.selectedFiles()[0]
                if filename:
                    self.filter_path = filename

        self.load_combo()

    def on_delete(self):
        filename = self.combo.currentData()
        if filename == None:
            self.model.clear()
            return

        msgBox = QMessageBox()
        msgBox.setText(f"Remove {filename}")
        msgBox.setInformativeText("Do you want to remove this file ?")
        msgBox.setStandardButtons(QMessageBox.Save | QMessageBox.Discard)
        msgBox.setDefaultButton(QMessageBox.Save)

        if msgBox.exec():
            if os.path.exists(filename):
                os.remove(filename)
                self.model.clear()
                self.load_combo()

    def on_combo_changed(self, index):
        """Called when a new item is selected in the ComboBox (programatically or not)"""
        filename = self.combo.itemData(index)
        print(filename)

        if filename:
            self.combo.setToolTip(filename)
            self.model.from_json(filename)
            self.on_filters_changed()
            self._update_view_geometry()
        # self.on_filters_changed()

    def _update_view_geometry(self):
        """Set column Spanned to True for all Logic Item

        Allow Logic Item Editor to take all the space inside the row
        """
        self.view.expandAll()

        # self.view.horizontalHeader().setSectionResizeMode(3, QHeaderView.ResizeToContents)
        # self.view.horizontalHeader().setSectionResizeMode(2, QHeaderView.ResizeToContents)

        # for index in self.model.match(
        #     self.model.index(0, 0),
        #     FilterModel.TypeRole,
        #     FilterItem.LOGIC_TYPE,
        #     -1,
        #     Qt.MatchRecursive,
        # ):
        #     self.view.setFirstColumnSpanned(0, index.parent(), True)

    def on_add_condition(self):
        """Add new condition item

        - Add condition item to the current selected operator
        - Or add new operator and new condition item on a new filter
        """
        index = self.view.currentIndex()

        if index.isValid():
            if self.model.item(index).type == FilterItem.LOGIC_TYPE:
                # Add condition item to existing logic operator
                self.model.add_condition_item(parent=index)
        else:
            if self.model.rowCount() == 0:
                # Full new logic operator and condition item
                self.model.add_logic_item(parent=QModelIndex())
                gpindex = self.model.index(0, 0, QModelIndex())
                self.model.add_condition_item(parent=gpindex)

        self._update_view_geometry()
        self.refresh_buttons()

    def on_open_condition_dialog(self):
        """Open the condition creation dialog
        TODO: not used anymore
        """
        dialog = FieldDialog(conn=self.conn, parent=self)
        if dialog.exec_() == dialog.Accepted:
            cond = dialog.get_condition()
            index = self.view.currentIndex()
            if index:
                self.model.add_condition_item(parent=index, value=cond)

    # def on_delete_item(self):
    #     """Delete current item

    #     - if filter is saved => delete filter
    #     - if filter is not saved => clear
    #     """
    #     ret = QMessageBox.question(
    #         self,
    #         self.tr("Filter deletion"),
    #         self.tr("Do you want to remove this filter?"),
    #         QMessageBox.Yes | QMessageBox.No,
    #     )

    #     if ret == QMessageBox.Yes:
    #         current_index = self.combo.currentIndex()
    #         if current_index == 0:
    #             # Not saved filter
    #             self.model.clear()
    #         else:
    #             # Saved filter
    #             # Delete from app settings
    #             self.settings.beginGroup("plugins/filters_editor/filters")
    #             self.settings.remove(self.combo.currentText())
    #             self.settings.endGroup()

    #             # Delete from combobox
    #             self.combo.removeItem(current_index)

    #         self.refresh_buttons()

    def on_selection_changed(self):
        """Enable/Disable add button depending item type

        Notes:
            Disable Add button on CONDITION_TYPE
        """
        self.refresh_buttons()

    # def contextMenuEvent(self, event: QContextMenuEvent):

    #     pos = self.view.viewport().mapFromGlobal(event.globalPos())
    #     index = self.view.indexAt(pos)

    #     if index.isValid():
    #         menu = QMenu(self)

    #         item = self.model.item(index)
    #         if item.type == FilterItem.LOGIC_TYPE:
    #             menu.addAction(self.tr("Add condition"), self.on_add_condition)
    #             menu.addAction(self.tr("Add subfilter"), self.on_add_logic)

    #         menu.exec_(event.globalPos())


if __name__ == "__main__":

    app = QApplication(sys.argv)
    app.setStyle("fusion")

    style.dark(app)

    from cutevariant.core.importer import import_reader
    from cutevariant.core.reader import FakeReader
    import cutevariant.commons as cm
    from cutevariant.gui.ficon import FIcon, setFontPath

    setFontPath(cm.FONT_FILE)

    conn = sql.get_sql_connection(":memory:")
    import_reader(conn, FakeReader())

    data = {
        "AND": [
            {"field": "gene", "operator": "=", "value": "chr12"},
            {"field": "gene", "operator": "=", "value": "chr12"},
            {"field": "gene", "operator": "=", "value": "chr12"},
            {"field": "gene", "operator": "=", "value": "chr12"},
            {"field": "gene", "operator": "=", "value": "chr12"},
            {"field": "gene", "operator": "=", "value": "chr12"},
            {
                "AND": [
                    {"field": "gene", "operator": "=", "value": "chr12"},
                    {"field": "gene", "operator": "=", "value": "chr12"},
                    {
                        "AND": [
                            {"field": "gene", "operator": "=", "value": "chr12"},
                            {"field": "gene", "operator": "=", "value": "chr12"},
                            {
                                "AND": [
                                    {
                                        "field": "gene",
                                        "operator": "=",
                                        "value": "chr12",
                                    },
                                    {
                                        "field": "gene",
                                        "operator": "=",
                                        "value": "chr12",
                                    },
                                ]
                            },
                        ]
                    },
                ]
            },
            {"field": "gene", "operator": "=", "value": "chr12"},
            {"field": "gene", "operator": "=", "value": "chr12"},
        ]
    }

    view = FiltersEditorWidget()
    view.model.conn = conn
    view.model.load(data)

    view._update_view_geometry()
    view.show()

    # view = QTreeView()
    # view.setEditTriggers(QAbstractItemView.DoubleClicked)
    # view.setAlternatingRowColors(True)
    # view.setUniformRowHeights(True)
    # view.setModel(model)
    # view.setAcceptDrops(True)
    # view.setDragEnabled(True)
    # view.setDropIndicatorShown(True)
    # view.setSelectionBehavior(QAbstractItemView.SelectRows)
    # view.setDragDropMode(QAbstractItemView.InternalMove)

    # view.setFirstColumnSpanned(0, QModelIndex(), True)
    # view.resize(500, 500)
    # view.show()
    # view.expandAll()

    app.exec_()



import os
import csv

import cutevariant.commons as cm

LOGGER = cm.logger()


class PedReader:
    r"""PED \*.tfam file (PLINK sample information file) parser

    Data has the same structure of a tfam file object
    https://www.cog-genomics.org/plink/1.9/formats#fam
    http://zzz.bwh.harvard.edu/plink/data.shtml

    Format description:
        A tabular/white space separated text file with no header line, and one
        line per sample with the following six fields:

        - Family ID (str): ('FID'); ex: "Valse"
        - Within-family ID (str): ('IID'; cannot be '0'); ex: "Emmanuel"
        - Within-family ID of father (str): ('0' if father isn't in dataset)
        - Within-family ID of mother (str): ('0' if mother isn't in dataset)
        - Sex code (int): ('1' = male, '2' = female, '0' or other = unknown)
        - Phenotype value (int): ('1' = control, '2' = case, '-9'/'0'/non-numeric = missing data if case/control)
            We expect '0' for this last one.

    Notes:
        Comments are lines starting with a # character.
        The rest of that line will be ignored.

    Recall, DB structure:
        - id INTEGER PRIMARY KEY ASC,
        - name TEXT,
        - family_id TEXT DEFAULT 'fam',
        - father_id INTEGER DEFAULT 0,
        - mother_id INTEGER DEFAULT 0,
        - sex INTEGER DEFAULT 0,
        - phenotype INTEGER DEFAULT 0

    TODO:
        Detect if an individual has unknown sex but appears in mother_id or
        father_id of a family.
    """

    def __init__(self, filepath: str, samples: dict, raw_samples=True, *args, **kwargs):
        """

        Args:
            filepath (str): PED filepath
            samples (dict): Database current samples. Each key is a field
                of the table named "samples". We use this dict to apply unicity
                constraints over family_id vs individual_id, father_id and
                mother_id.
                Must be not empty if `raw_samples` is `False`.
            raw_samples (boolean):
                - If raw_samples are activated (default), samples are lists of fields.
                - If raw_samples are not activated, samples are dict of fields
                  ready to be inserted in the database according to the given samples.

        """
        assert os.path.isfile(filepath)
        assert (
            samples and not raw_samples
        ) or raw_samples, (
            "If raw_samples are deactivated, database samples must be given"
        )

        self.filepath = filepath
        self.samples = samples
        self.raw_samples = raw_samples

    def __iter__(self):
        """Generator on PED file

        PED file is opened as a tabulated or white space separated file.
        """
        with open(self.filepath, "r") as stream:
            # Sniff dialect of CSV file
            dialect = csv.Sniffer().sniff(stream.read(10000), delimiters="\t ")
            stream.seek(0)

            reader = csv.reader(
                (row for row in stream if not row.startswith("#")),  # Remove comments
                dialect,
            )

            yield from self.get_samples(reader)

    def get_samples(self, reader):
        """Yield samples from PED file

        Notes:
            The following problems with samples are detected:
                - Not digit sex/phenotype: Exception raised
                - Sex/phenotype not expected (0,1,2): Individual skipped
                - Unknown individual_id: Individual skipped
                - If samples are given, not found family_id, individual_id couple:
                  Individual skipped.
                  i.e.: And individual not already in DB is skipped.
                - If samples are given, not found family_id, father_id/mother_id:
                  Added as unknown.

        Returns:
            (generator[dict/list]): Generator of samples.
                - If raw_samples are activated (default), samples are lists of fields.
                - If raw_samples are not activated, samples are dict of fields
                  ready to be inserted in the database according to the given samples.

            Example::

                `[family_id, individual_id, father_id, mother_id, sex, phenotype]`
                Or
                `["id": _, "family_id": _, "father_id": _, "mother_id": _, "sex": _, "phenotype": ]`
        """
        samples_mapping = {sample["name"]: sample["id"] for sample in self.samples}

        for index, line in enumerate(reader, 1):
            if len(line) < 6:
                LOGGER.error(
                    "PED file conformity line <%s>; too few fields; expected at least 6",
                    index,
                )
                continue

            # Extract and validate data
            family_id = line[0]
            individual_id = line[1]
            father_id = line[2]
            mother_id = line[3]
            sex = int(line[4]) if line[4].isdigit() else 0
            phenotype = int(line[5]) if line[5].isdigit() else 0

            if sex not in (0, 1, 2):
                LOGGER.error(
                    "PED file conformity line <%s>; sex code <%s> not expected",
                    index,
                    sex,
                )
                continue

            if phenotype not in (0, 1, 2):
                LOGGER.error(
                    "PED file conformity line <%s>; phenotype code <%s> not expected",
                    index,
                    sex,
                )
                continue

            if individual_id == "0":
                LOGGER.error(
                    "PED file conformity line <%s>; Within-family ID/Sample name can't be '0'",
                    index,
                )
                continue

            # Test presence of sample in DB
            # individual_key = (family_id, individual_id)
            # if samples_mapping and individual_key not in samples_mapping.keys():
            #     LOGGER.error(
            #         "PED file conformity line <%s>; sample (family_id, individual_id):"
            #         "<%s, %s> not found in database",
            #         index, family_id, individual_id,
            #     )
            #     continue

            # Test presence of parents in DB

            if self.raw_samples:
                new_sample = [
                    family_id,
                    individual_id,
                    father_id,
                    mother_id,
                    sex,
                    phenotype,
                ]

            else:

                new_sample = {
                    "id": samples_mapping[individual_id],  # Get DB sample id
                    "family_id": family_id,
                    "father_id": samples_mapping.get(father_id, 0),
                    "mother_id": samples_mapping.get(mother_id, 0),
                    "sex": sex,
                    "phenotype": phenotype,
                }

            # print(new_sample)
            yield new_sample



"""A place to store style rules for the GUI"""
from cutevariant.commons import DIR_STYLES
from PySide2.QtGui import QPalette, QColor


CLASSIFICATION = {
    0: {"name": "Unclassified", "icon": 0xF03A1, "color": "lightgray"},
    1: {"name": "Benin", "icon": 0xF03A4, "color": "#71e096"},
    2: {"name": "Likely benin", "icon": 0xF03A7, "color": "#71e096"},
    3: {
        "name": "Variant of uncertain significance",
        "icon": 0xF03AA,
        "color": "#f5a26f",
    },
    4: {"name": "Likely pathogen", "icon": 0xF03AD, "color": "#ed6d79"},
    5: {"name": "Pathogen", "icon": 0xF03B1, "color": "#ed6d79"},
}


GENOTYPE = {
    -1: {"name": "Unknown genotype", "icon": 0xF10D3},
    0: {"name": "Homozygous wild", "icon": 0xF0766},
    1: {"name": "Heterozygous", "icon": 0xF0AA1},
    2: {"name": "Homozygous muted", "icon": 0xF0AA5},
}


FIELD_TYPE = {
    "float": {"name": "floating ", "icon": 0xF03A0, "color": "#71e096"},
    "int": {"name": "integer", "icon": 0xF03A0, "color": "#f5a26f"},
    "str": {"name": "text", "icon": 0xF100D, "color": "#ed6d79"},
    "bool": {"name": "boolean", "icon": 0xF023B, "color": "#949cdf"},
}


GENE_COLOR = "#F5A26F"
WARNING_BACKGROUND_COLOR = "#FFCCBA"
WARNING_TEXT_COLOR = "#D73705"

DARK_COLOR = {
    "darkpurple": "#40375C",
    "purple": "#5A4F7C",
    "red": "#F14235",
    "yellow": "#F5A623",
    "green": "#7BBB44",
}


# Sequence ontology colors
SO_COLOR = {
    "stop": "#808080",
    "utr": "#911EB4",
    "splice": "#F58231",
    "intron": "#0082C8",
    "intergenic": "#3CB44B",
    "frameshift": "#E6194B",
    "missense": "#F032E6",
}


def bright(app):
    """Mock function to don't apply any style to the Qt application instance.

    TODO: Find a way to reset properly and dynamically the effects of `dark()`
    and put it here.

    .. note:: Called my __main__ on startup and not by StyleSettingsWidget for
    the moment (see TODO).
    """
    pass


def dark(app):
    """Apply Dark Theme to the Qt application instance.
        Args:
            app (QApplication): QApplication instance.
    """

    darkPalette = QPalette()

    # base
    darkPalette.setColor(QPalette.WindowText, QColor(180, 180, 180))
    darkPalette.setColor(QPalette.Button, QColor(53, 53, 53))
    darkPalette.setColor(QPalette.Light, QColor(180, 180, 180))
    darkPalette.setColor(QPalette.Midlight, QColor(90, 90, 90))
    darkPalette.setColor(QPalette.Dark, QColor(35, 35, 35))
    darkPalette.setColor(QPalette.Text, QColor(180, 180, 180))
    darkPalette.setColor(QPalette.BrightText, QColor(200, 200, 200))
    darkPalette.setColor(QPalette.ButtonText, QColor(180, 180, 180))
    darkPalette.setColor(QPalette.Base, QColor(42, 42, 42))
    darkPalette.setColor(QPalette.Window, QColor(53, 53, 53))
    darkPalette.setColor(QPalette.Shadow, QColor(20, 20, 20))
    darkPalette.setColor(QPalette.Highlight, QColor(42, 130, 218))
    darkPalette.setColor(QPalette.HighlightedText, QColor(180, 180, 180))
    darkPalette.setColor(QPalette.Link, QColor(56, 252, 196))
    darkPalette.setColor(QPalette.AlternateBase, QColor(66, 66, 66))
    darkPalette.setColor(QPalette.ToolTipBase, QColor(53, 53, 53))
    darkPalette.setColor(QPalette.ToolTipText, QColor(180, 180, 180))

    # disabled
    darkPalette.setColor(QPalette.Disabled, QPalette.WindowText, QColor(127, 127, 127))
    darkPalette.setColor(QPalette.Disabled, QPalette.Text, QColor(127, 127, 127))
    darkPalette.setColor(QPalette.Disabled, QPalette.ButtonText, QColor(127, 127, 127))
    darkPalette.setColor(QPalette.Disabled, QPalette.Highlight, QColor(80, 80, 80))
    darkPalette.setColor(
        QPalette.Disabled, QPalette.HighlightedText, QColor(127, 127, 127)
    )

    app.setPalette(darkPalette)

    # _apply_base_theme(app)
    with open(DIR_STYLES + "dark.qss", "r") as file:
        app.setStyleSheet(file.read())


def apply_frameless_style(widget):
    """Apply frameless style to the given widget

    TODO: What this style is supposed to do ?
    """
    with open(DIR_STYLES + "frameless.qss", "r") as file:
        widget.setStyleSheet(file.read())



from PySide2.QtWidgets import (
    QTableView,
    QAbstractItemView,
    QApplication,
    QWidget,
    QVBoxLayout,
    QLineEdit,
    QMenu,
    QAction,
)
from PySide2.QtCore import QAbstractTableModel, QModelIndex, QSortFilterProxyModel, Qt
from PySide2.QtGui import (
    QFont,
    QContextMenuEvent,
    QColor,
    QKeySequence,
    QPaintEvent,
    QPainter,
)


class DictModel(QAbstractTableModel):
    def __init__(self, parent=None):
        super().__init__(parent)
        self._raw_data = []
        self._headers = ["Key", "Value"]

    def columnCount(self, parent: QModelIndex = QModelIndex()) -> int:
        return 2

    def rowCount(self, parent=QModelIndex()) -> int:
        return len(self._raw_data)

    def flags(self, index: QModelIndex) -> Qt.ItemFlags:

        if index.column() == 0:
            return Qt.ItemIsSelectable | Qt.ItemIsEnabled

        if index.column() == 1:
            return Qt.ItemIsSelectable | Qt.ItemIsEnabled

    def data(self, index: QModelIndex, role: Qt.ItemDataRole = Qt.DisplayRole):

        if not index.isValid():
            return None

        if role == Qt.DisplayRole or role == Qt.ToolTipRole:
            return self.value(index)

        if role == Qt.FontRole:
            value = self.value(index)
            font = QFont()
            if index.column() == 0:
                font.setBold(True)
            if index.column() == 1 and value == "NULL":
                font.setItalic(True)
            return font

        if role == Qt.TextColorRole:
            value = self.value(index)
            if index.column() == 1 and value == "NULL":
                return QColor("lightgray")

        return None

    def value(self, index: QModelIndex):
        if index.isValid():
            value = self._raw_data[index.row()][index.column()]
            if value is None:
                return "NULL"
            else:
                return value
        else:
            return None

    def headerData(
        self, section: int, orientation: Qt.Orientation, role: Qt.ItemDataRole
    ):
        if orientation == Qt.Horizontal and role == Qt.DisplayRole:
            return self._headers[section]

        return None

    def set_dict(self, data: dict):

        if data:
            self.beginResetModel()
            self._raw_data = [(k, v) for k, v in data.items()]
            self.endResetModel()

    def clear(self):
        self.set_dict({})


class DictWidget(QWidget):
    def __init__(self, parent=None):
        super().__init__()

        self.view = QTableView()
        self.model = DictModel()
        self.proxy_model = QSortFilterProxyModel()
        self.search_bar = QLineEdit()
        self._show_loading = False

        self.proxy_model.setSourceModel(self.model)

        self.view.setModel(self.proxy_model)
        self.view.setAlternatingRowColors(True)
        self.view.horizontalHeader().setStretchLastSection(True)
        self.view.setSelectionMode(QAbstractItemView.SingleSelection)
        self.view.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.view.setSortingEnabled(True)
        self.view.verticalHeader().hide()

        self.search_bar.textChanged.connect(self.proxy_model.setFilterRegularExpression)
        self.search_bar.setVisible(False)

        self._show_search_action = QAction("show search bar")
        self._show_search_action.setCheckable(True)
        self._show_search_action.setShortcutContext(Qt.WidgetShortcut)
        self._show_search_action.setShortcut(QKeySequence.Find)
        self._show_search_action.triggered.connect(self._on_show_search)

        self._close_search_action = QAction()
        self._close_search_action.setShortcut(QKeySequence(Qt.Key_Escape))
        self._close_search_action.setShortcutContext(Qt.WidgetShortcut)
        self._close_search_action.triggered.connect(self._on_close_search)

        self.view.addAction(self._show_search_action)
        self.search_bar.addAction(self._close_search_action)

        _layout = QVBoxLayout()
        _layout.addWidget(self.view)
        _layout.addWidget(self.search_bar)
        _layout.setContentsMargins(0, 0, 0, 0)

        self.setLayout(_layout)
        print("init")

    def set_dict(self, data: dict):

        self.model.set_dict(data)

    def _on_show_search(self):
        self.search_bar.setVisible(True)
        self.search_bar.setFocus(Qt.ShortcutFocusReason)

    def _on_close_search(self):
        self.search_bar.hide()
        self.search_bar.clear()
        self.view.setFocus(Qt.ShortcutFocusReason)

    def set_header_visible(self, visible=True):
        self.view.horizontalHeader().setVisible(visible)

    def clear(self):
        self.model.clear()

    def paintEvent(self, event: QPaintEvent):

        if self._show_loading:
            painter = QPainter(self)
            painter.drawText(self.rect(), Qt.AlignCenter, self.tr("Loading ..."))
        else:
            super().paintEvent(event)

    def set_loading(self, show=True):
        self._show_loading = True
        self.view.setVisible(not show)
        self.update()


if __name__ == "__main__":
    import sys

    app = QApplication(sys.argv)

    w = DictWidget()
    w.set_dict({"name": "sacha", "age": 13, "sexe": True})

    w.show()
    app.exec_()



"""Exposes VqlEditor class used in the MainWindow to show VQL statements.
VqlEditor uses:
    - VqlSyntaxHighlighter as a syntax highlighter
    - VqlEdit for the support of autocompletion

Warnings:
    When you add a VQL command in the language, please update
    :meth:`VqlEditorWidget.run_vql` method in consequence. Otherwise no command
    will be executed.
"""
# Standard imports
from textx.exceptions import TextXSyntaxError
import sqlite3

# Qt imports
from PySide2.QtCore import Qt, QStringListModel
from PySide2.QtWidgets import (
    QToolBar,
    QCompleter,
    QVBoxLayout,
    QLabel,
    QFrame,
    QApplication,
)
from PySide2.QtGui import QKeySequence

# Custom imports
from cutevariant.core import vql, sql
from cutevariant.gui import style, plugin, FIcon
from cutevariant.core.vql import VQLSyntaxError
from cutevariant.core import command
from cutevariant.core.querybuilder import build_vql_query
from cutevariant.gui.widgets import CodeEdit
from cutevariant.commons import logger

LOGGER = logger()


class VqlEditorWidget(plugin.PluginWidget):
    """Exposed class to manage VQL/SQL queries from the mainwindow"""

    LOCATION = plugin.FOOTER_LOCATION
    ENABLE = True

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setWindowTitle(self.tr("VQL Editor"))

        # Top toolbar
        self.top_bar = QToolBar()
        self.top_bar.setToolButtonStyle(Qt.ToolButtonTextBesideIcon)
        self.run_action = self.top_bar.addAction(
            FIcon(0xF040A), self.tr("Run"), self.run_vql
        )
        self.run_action.setShortcuts([Qt.CTRL + Qt.Key_R, QKeySequence.Refresh])
        self.run_action.setToolTip(
            self.tr("Run VQL query (%s)" % self.run_action.shortcut().toString())
        )

        # Syntax highlighter and autocompletion
        self.text_edit = CodeEdit()
        # Error handling
        self.log_edit = QLabel()
        self.log_edit.setMinimumHeight(40)
        self.log_edit.setStyleSheet(
            "QWidget{{background-color:'{}'; color:'{}'}}".format(
                style.WARNING_BACKGROUND_COLOR, style.WARNING_TEXT_COLOR
            )
        )
        self.log_edit.hide()
        self.log_edit.setFrameStyle(QFrame.StyledPanel | QFrame.Raised)

        main_layout = QVBoxLayout()
        main_layout.addWidget(self.top_bar)
        main_layout.addWidget(self.text_edit)
        main_layout.addWidget(self.log_edit)
        main_layout.setContentsMargins(0, 0, 0, 0)
        main_layout.setSpacing(0)
        self.setLayout(main_layout)

    def on_open_project(self, conn: sqlite3.Connection):
        """overrided from PluginWidget : Do not call this methods

        Args:
            conn (sqlite3.Connection): sqlite3 connection
        """
        self.conn = conn
        self._fill_completer()

        self.on_refresh()

    def on_refresh(self):
        """overrided from PluginWidget"""

        vql_query = build_vql_query(
            self.mainwindow.state.fields,
            self.mainwindow.state.source,
            self.mainwindow.state.filters,
            self.mainwindow.state.group_by,
            self.mainwindow.state.having,
        )

        self.set_vql(vql_query)

    def set_vql(self, text: str):
        """Set vql source code without executed

        Args:
            text (str): VQL query
        """
        self.text_edit.blockSignals(True)
        self.text_edit.setPlainText(text)
        self.text_edit.blockSignals(False)

    def _fill_completer(self):
        """Create Completer with his model

        Fill the model with the SQL keywords and database fields
        """
        # preload samples , selection and wordset
        samples = [i["name"] for i in sql.get_samples(self.conn)]
        selections = [i["name"] for i in sql.get_selections(self.conn)]
        wordsets = [i["name"] for i in sql.get_wordsets(self.conn)]

        # keywords = []
        self.text_edit.completer.model.clear()
        self.text_edit.completer.model.beginResetModel()

        # register keywords
        for keyword in self.text_edit.syntax.sql_keywords:
            self.text_edit.completer.model.add_item(
                keyword, "VQL keywords", FIcon(0xF0169), "#f6ecf0"
            )

        for selection in selections:
            self.text_edit.completer.model.add_item(
                selection, "Source table", FIcon(0xF04EB), "#f6ecf0"
            )

        for wordset in wordsets:
            self.text_edit.completer.model.add_item(
                f"WORDSET['{wordset}']", "WORDSET", FIcon(0xF04EB), "#f6ecf0"
            )

        for field in sql.get_fields(self.conn):
            name = field["name"]
            description = "<b>{}</b> ({}) from {} <br/><br/> {}".format(
                field["name"], field["type"], field["category"], field["description"]
            )
            color = style.FIELD_TYPE.get(field["type"], "str")["color"]
            icon = FIcon(style.FIELD_TYPE.get(field["type"], "str")["icon"], "white")

            if field["category"] == "variants" or field["category"] == "annotations":
                self.text_edit.completer.model.add_item(name, description, icon, color)

            if field["category"] == "samples":
                # Overwrite name
                for sample in samples:
                    name = "sample['{}'].{}".format(sample, field["name"])
                    description = "<b>{}</b> ({}) from {} {} <br/><br/> {}".format(
                        field["name"],
                        field["type"],
                        field["category"],
                        sample,
                        field["description"],
                    )
                    self.text_edit.completer.model.add_item(
                        name, description, icon, color
                    )

        self.text_edit.completer.model.endResetModel()

        # if field["category"] == "samples":
        #     for sample in samples:
        #         keywords.append("sample['{}'].{}".format(sample, field["name"]))
        # else:
        #     keywords.append(field["name"])

    def check_vql(self) -> bool:
        """Check VQL statement; return True if OK, False when an error occurs

        Notes:
            This function also sets the error message to the bottom of the view.

        Returns:
            bool: Status of VQL query (True if valid, False otherwise).
        """
        try:
            self.log_edit.hide()
            _ = [i for i in vql.parse_vql(self.text_edit.toPlainText())]
        except (TextXSyntaxError, VQLSyntaxError) as e:
            # Show the error message on the ui
            # Available attributes: e.message, e.line, e.col
            self.set_message(
                "%s: %s, col: %d" % (e.__class__.__name__, e.message, e.col)
            )
            return False
        return True

    def run_vql(self):
        """Execute VQL code

        Suported commands and the plugins that need to be refreshed in consequence:
            - select_cmd: main ui (all plugins in fact)
            - count_cmd: *not supported*
            - drop_cmd: selections & wordsets
            - create_cmd: selections
            - set_cmd: selections
            - bed_cmd: selections
            - show_cmd: *not supported*
            - import_cmd: wordsets
        """
        # Check VQL syntax first
        if not self.check_vql():
            return

        for cmd in vql.parse_vql(self.text_edit.toPlainText()):

            LOGGER.debug("VQL command %s", cmd)
            cmd_type = cmd["cmd"]

            # If command is a select kind
            if cmd_type == "select_cmd":
                # => Command will be executed in different widgets (variant_view)
                # /!\ VQL Editor will not check SQL validity of the command
                # columns from variant table
                self.mainwindow.state.fields = cmd["fields"]
                # name of the variant selection
                self.mainwindow.state.source = cmd["source"]
                self.mainwindow.state.filters = cmd["filters"]
                self.mainwindow.state.group_by = cmd["group_by"]
                self.mainwindow.state.having = cmd["having"]
                # Refresh all plugins
                self.mainwindow.refresh_plugins(sender=self)
                continue

            try:
                # Check SQL validity of selections related commands
                command.create_command_from_obj(self.conn, cmd)()
            except (sqlite3.DatabaseError, VQLSyntaxError) as e:
                # Display errors in VQL editor
                self.set_message(str(e))
                LOGGER.exception(e)
                continue

            # Selections related commands
            if cmd_type in ("create_cmd", "set_cmd", "bed_cmd"):
                # refresh source editor plugin for selections
                self.mainwindow.refresh_plugin("source_editor")
                continue

            if cmd_type == "drop_cmd":
                # refresh source editor plugin for selections
                self.mainwindow.refresh_plugin("source_editor")
                # refresh wordset plugin
                self.mainwindow.refresh_plugin("word_set")

            if cmd_type == "import_cmd":
                # refresh wordset plugin
                self.mainwindow.refresh_plugin("word_set")

    def set_message(self, message: str):
        """Show message error at the bottom of the view

        Args:
            message (str): Error message
        """
        if self.log_edit.isHidden():
            self.log_edit.show()

        icon_64 = FIcon(0xF0027, style.WARNING_TEXT_COLOR).to_base64(18, 18)

        self.log_edit.setText(
            """<div height=100%>
            <img src="data:image/png;base64,{}" align="left"/>
             <span> {} </span>
            </div>""".format(
                icon_64, message
            )
        )


if __name__ == "__main__":

    import sys

    app = QApplication(sys.argv)

    conn = sqlite3.connect("examples/test.db")

    view = VqlEditorWidget()
    view.show()

    app.exec_()



# Standard imports
import sqlite3

# Qt imports
from PySide2.QtWidgets import (
    QVBoxLayout,
    QTableView,
    QDialogButtonBox,
    QAbstractItemView,
    QHeaderView,
    QTabWidget,
    QStatusBar,
)
from PySide2.QtCore import Qt, QAbstractTableModel, QModelIndex, QThreadPool

# Custom imports
from cutevariant.gui.plugin import PluginDialog
from cutevariant.gui.sql_thread import SqlThread
from cutevariant.gui.widgets import DictWidget
from cutevariant.core import sql


# SQL functions
def get_variant_count(conn: sqlite3.Connection):
    return conn.execute(
        "SELECT `count` FROM selections WHERE name = 'variants'"
    ).fetchone()["count"]


def get_variant_transition(conn: sqlite3.Connection):
    return conn.execute(
        """SELECT COUNT(*) AS `count` FROM variants
        WHERE (ref == 'A' AND alt == 'G')
        OR (ref == 'G' AND alt == 'A')
        OR (ref == 'C' AND alt == 'T')
        OR (ref == 'T' AND alt == 'C')"""
    ).fetchone()["count"]


def get_variant_transversion(conn: sqlite3.Connection):
    return conn.execute(
        """SELECT COUNT(*) AS `count` FROM variants
        WHERE (ref == 'A' AND alt == 'C')
        OR (ref == 'C' AND alt == 'A')
        OR (ref == 'G' AND alt == 'T')
        OR (ref == 'T' AND alt == 'G')
        OR (ref == 'G' AND alt == 'C')
        OR (ref == 'C' AND alt == 'G')
        OR (ref == 'A' AND alt == 'T')
        OR (ref == 'T' AND alt == 'A')"""
    ).fetchone()["count"]


def get_sample_count(conn: sqlite3.Connection):
    return conn.execute("SELECT COUNT(*) AS `count` FROM samples").fetchone()["count"]


def get_snp_count(conn: sqlite3.Connection):
    """Get the number of variants that are SNP

    Notes:
        This query is currently not covered by an index.
    """
    return conn.execute(
        "SELECT COUNT(*) AS `count` FROM variants WHERE is_snp = 1"
    ).fetchone()["count"]


def get_indel_count(conn: sqlite3.Connection):
    """Get the number of variants that are SNP

    Notes:
        This query is currently not covered by an index.
    """
    return conn.execute(
        "SELECT COUNT(*) AS `count` FROM variants WHERE is_indel = 1"
    ).fetchone()["count"]


def get_gene_counts(conn: sqlite3.Connection):
    """ Get the number of variant per genes """
    results = {}
    for record in conn.execute(
        "SELECT gene, COUNT(*) as 'count' FROM annotations GROUP BY gene ORDER by count DESC LIMIT 1,100"
    ):
        results[record["gene"]] = record["count"]

    return results


class MetricsDialog(PluginDialog):

    ENABLE = True

    def __init__(self, conn=None, parent=None):
        super().__init__(parent)
        self.conn = conn

        self.tab_widget = QTabWidget()
        self.status_bar = QStatusBar()

        self.meta_view = DictWidget()
        self.stat_view = DictWidget()
        self.ann_view = DictWidget()

        self.tab_widget.addTab(self.meta_view, "Metadata")
        self.tab_widget.addTab(self.stat_view, "Variants")
        self.tab_widget.addTab(self.ann_view, "Annotations")

        self.buttons = QDialogButtonBox(QDialogButtonBox.Ok)

        self.buttons.accepted.connect(self.accept)

        self.setWindowTitle(self.tr("Project metrics"))

        v_layout = QVBoxLayout()
        v_layout.addWidget(self.tab_widget)
        v_layout.addWidget(self.status_bar)
        v_layout.addWidget(self.buttons)
        v_layout.setSpacing(0)
        self.setLayout(v_layout)

        self.resize(640, 480)
        # Async stuff
        self.metric_thread = None
        self.populate()

    def populate(self):
        """Async implementation to populate the view

        Notes:
            When closing the dialog window, the thread is not stopped.
        """

        def compute_metrics(conn):
            """Async function"""

            meta_data = sql.get_metadatas(conn)

            stats_data = {
                "Variant count": get_variant_count(conn),
                "Snp count": get_snp_count(conn),
                "Indel count": get_indel_count(conn),
                "Transition count": get_variant_transition(conn),
                "Transversion count": get_variant_transversion(conn),
                "Sample count": get_sample_count(conn),
            }

            stats_data["Tr/tv ratio"] = round(
                stats_data["Transition count"] / stats_data["Transversion count"], 2
            )

            if sql.table_exists(conn, "annotations"):
                genes_data = get_gene_counts(conn)
            else:
                gene_data = {}

            return meta_data, stats_data, genes_data

        self.status_bar.showMessage("Loading ...")
        self.metric_thread = SqlThread(self.conn, compute_metrics)
        self.metric_thread.result_ready.connect(self.loaded)
        self.metric_thread.start()

    def loaded(self):
        """Called at the end of the thread and populate data"""
        meta_data, stats_data, genes_data = self.metric_thread.results

        self.stat_view.set_dict(stats_data)
        self.meta_view.set_dict(meta_data)
        self.ann_view.set_dict(genes_data)
        self.status_bar.showMessage("")


if __name__ == "__main__":
    from PySide2.QtWidgets import QApplication
    import sys

    app = QApplication(sys.argv)
    conn = sql.get_sql_connection("test.db")
    dialog = MetricsDialog(conn=conn)
    app.exec_()



"""File importer and project creation"""
# Custom imports
from cutevariant import __version__
from .reader.abstractreader import AbstractReader
from .readerfactory import create_reader
from cutevariant.core.reader.pedreader import PedReader
from .sql import (
    create_project,
    create_table_metadatas,
    insert_many_metadatas,
    create_table_fields,
    create_table_annotations,
    create_table_variants,
    create_table_samples,
    create_table_selections,
    create_table_wordsets,
    insert_many_samples,
    get_samples,
    insert_many_fields,
    async_insert_many_variants,
    create_indexes,
    update_sample,
)
from cutevariant.commons import logger

LOGGER = logger()


def async_import_reader(
    conn, reader: AbstractReader, pedfile=None, ignored_fields=set(), project={}
):
    """Import data via the given reader into a SQLite database via the given connection

    :param conn: sqlite connection
    :param reader: must be a AbstractReader base class
    :param pedfile: PED file path
    :param project: The reference genome and the name of the project.
        Keys have to be at least "reference" and "project_name".
    :type project: <dict>
    :return: yield progression and message
    :rtype: <generator <int>, <str>>
    """
    # Create project
    yield 0, f"Importing data with {reader}"
    create_project(
        conn,
        name=project.get("project_name", "UKN"),
        reference=project.get("reference", "UKN"),
    )

    # Create metadatas
    create_table_metadatas(conn)
    metadatas = reader.get_metadatas()
    # Database versioning
    metadatas["cutevariant_version"] = __version__
    insert_many_metadatas(conn, metadatas)

    yield 0, "Creating table shema..."

    # Setup ignored fields
    reader.ignored_fields = ignored_fields

    # Create table fields
    create_table_fields(conn)

    # Create annotations tables
    create_table_annotations(conn, reader.get_extra_fields_by_category("annotations"))

    # Create variants tables
    create_table_variants(conn, reader.get_extra_fields_by_category("variants"))

    # Create table samples
    create_table_samples(conn, reader.get_extra_fields_by_category("samples"))

    # Create selection
    create_table_selections(conn)

    # Create table sets
    create_table_wordsets(conn)

    # Insert samples
    yield 0, "Inserting samples..."
    insert_many_samples(conn, reader.get_samples())

    # Import PED file
    if pedfile:
        yield 0, f"Import pedfile {pedfile}"
        import_pedfile(conn, pedfile)

        # Compute control and cases samples
        samples = tuple(get_samples(conn))
        LOGGER.debug("Check found samples in DB after PED import: %s", samples)
        control_samples = [
            sample["name"] for sample in samples if sample["phenotype"] == 1
        ]
        case_samples = [
            sample["name"] for sample in samples if sample["phenotype"] == 2
        ]
        yield 0, "Compute phenotypes from samples:"
        yield 0, "- Found controls are: [" + ",".join(control_samples) + "]"
        yield 0, "- Found cases are: [" + ",".join(case_samples) + "]"
    else:
        control_samples = list()
        case_samples = list()

    # Insert fields
    yield 0, "Inserting fields..."
    insert_many_fields(conn, reader.get_extra_fields())

    # Insert variants, link them to annotations and samples
    yield 0, "Insertings variants..."
    variants = reader.get_extra_variants(control=control_samples, case=case_samples)
    yield from async_insert_many_variants(
        conn, variants, total_variant_count=reader.number_lines
    )

    # Create indexes
    yield 99, "Creating indexes..."
    create_indexes(conn)
    yield 100, "Indexes created."

    # session.add(Selection(name="favoris", description="favoris", count = 0))


def async_import_file(conn, filename, pedfile=None, ignored_fields=set(), project={}):
    """Import filename into SQLite database

    :param conn: sqlite connection
    :param filename: variant filename
    :param pedfile: PED file path
    :param project: The reference genome and the name of the project.
        Keys have to be at least "reference" and "project_name".
    :type project: <dict>
    :return: yield progression and message
    """
    # Context manager that wraps the given file and creates an apropriate reader
    with create_reader(filename) as reader:
        yield from async_import_reader(conn, reader, pedfile, ignored_fields, project)


def import_file(conn, filename, pedfile=None, ignored_fields=set(), project={}):
    """Wrapper for debugging purpose

    TODO: to be deleted
    """
    for progress, message in async_import_file(
        conn, filename, pedfile, ignored_fields, project
    ):
        # don't show message
        pass


def import_reader(conn, reader, pedfile=None, ignored_fields=set(), project={}):
    """Wrapper for debugging purpose

    TODO: to be deleted
    """
    for progress, message in async_import_reader(
        conn, reader, pedfile, ignored_fields, project
    ):
        # don't show message
        pass


def import_pedfile(conn, filename):
    r"""Import \*.tfam file (PLINK sample information file) into samples table

    See Also:
        :meth:`cutevariant.core.reader.pedreader.PedReader`
    """
    [
        update_sample(conn, sample)
        for sample in PedReader(filename, get_samples(conn), raw_samples=False)
    ]



# Standard imports
import importlib
import pkgutil
import os
import inspect

# Qt imports
from PySide2.QtCore import Qt, QModelIndex, QRect, QUrl
from PySide2.QtWidgets import QStyleOptionViewItem
from PySide2.QtGui import QIcon, QPainter, QFont, QPen, QColor


class Formatter(object):
    """Helper to customize cell style from QueryModel.
    You can set the font, background, foreground and decoration (QIcon)

    Class attributes:
        - DISPLAY_NAME: Name of the formatter displayed on the GUI.
    """

    DISPLAY_NAME = ""

    def __init__(self):
        super().__init__()

    def paint(
        self, painter: QPainter, option: QStyleOptionViewItem, index: QModelIndex
    ):
        painter.drawText(option.rect, option.displayAlignment, self.value(index))

    def field_name(self, index: QModelIndex):
        return index.model().headerData(index.column(), Qt.Horizontal)

    def value(self, index: QModelIndex):
        return index.data(Qt.DisplayRole)

    def draw_icon(self, painter: QPainter, rect: QRect, icon: QIcon):
        r = QRect(0, 0, 20, 20)
        r.moveCenter(rect.center())
        painter.drawPixmap(r, icon.pixmap(20, 20))

    def draw_url(self, painter: QPainter, rect: QRect, value: str):
        font = QFont()
        font.setUnderline(True)
        painter.setFont(font)
        painter.setPen(QPen(QColor("blue")))
        painter.drawText(rect, Qt.AlignCenter, value)


################################################################################


def find_formatters(path=None):
    """Find and return formatter classes from a directory

    Formatters must be found in `cutevariant/gui/formatters/`

    Keyword Arguments:
        path [str] -- the folder path where plugin are

    Returns:
        [generator [Formatter]] -- A Formatter class ready to be instantiated
    """

    # if path is None, return internal plugin path
    if path is None:
        formatter_path = os.path.join(os.path.dirname(__file__), "formatters")
    else:
        formatter_path = path

    for package in pkgutil.iter_modules([formatter_path]):
        package_path = os.path.join(formatter_path, package.name)
        spec = importlib.util.spec_from_file_location(
            package.name, package_path + ".py"
        )
        module = spec.loader.load_module()

        for name, obj in inspect.getmembers(module):
            # if the object is a class, whether built-in or created
            # => same as isinstance(obj, type)
            # + test parent class (Formatter)
            if inspect.isclass(obj) and "Formatter" in str(obj.__bases__):
                yield obj


if __name__ == "__main__":
    find_formatters()



from .vqleditor import VqlSyntaxHighlighter, VqlEditor  # deprecated

from .markdown_editor import MarkdownEditor

from .dict_widget import DictModel, DictWidget

from .codeedit import CodeEdit

from .file_edit import FileEdit



# Standard imports
from contextlib import contextmanager
import pathlib
import vcf

# Custom imports
from cutevariant.core.reader import VcfReader, CsvReader
import cutevariant.commons as cm


LOGGER = cm.logger()


def detect_vcf_annotation(filepath):
    """Return the name of the annotation parser to be used on the given file

    Called: In the importer and in the project wizard to display the detected
    annotations.

    :return: "vep", "snpeff", None
    """
    if cm.is_gz_file(filepath):
        # Open .gz files in binary mode (See #84)
        device = open(filepath, "rb")
    else:
        device = open(filepath, "r")

    std_reader = vcf.VCFReader(device)
    # print(std_reader.metadata)

    if "VEP" in std_reader.metadata:
        if "CSQ" in std_reader.infos:
            device.close()
            return "vep"

    if "SnpEffVersion" in std_reader.metadata:
        if "ANN" in std_reader.infos:
            device.close()
            return "snpeff"


@contextmanager
def create_reader(filepath):
    """Context manager that wraps the given file and return an accurate reader

    A detection of the file type is made as well as a detection of the
    annotations format if required.

    Filetypes and annotations parsers supported:

        - vcf.gz: snpeff, vep
        - vcf: snpeff, vep
        - csv, tsv, txt: vep
    """
    path = pathlib.Path(filepath)

    LOGGER.debug(
        "create_reader: PATH suffix %s, is_gz_file: %s",
        path.suffixes,
        cm.is_gz_file(filepath),
    )

    if ".vcf" in path.suffixes and ".gz" in path.suffixes:
        annotation_detected = detect_vcf_annotation(filepath)
        device = open(filepath, "rb")
        reader = VcfReader(device, annotation_parser=annotation_detected)
        yield reader
        device.close()
        return

    if ".vcf" in path.suffixes:
        annotation_detected = detect_vcf_annotation(filepath)
        device = open(filepath, "r")
        reader = VcfReader(device, annotation_parser=annotation_detected)
        yield reader
        device.close()
        return

    if {".tsv", ".csv", ".txt"} & set(path.suffixes):
        device = open(filepath, "r")
        reader = CsvReader(device)
        yield reader
        device.close()
        return

    raise Exception("create_reader:: Could not choose parser for this file.")



class State:
    """Store application state variable

    Settings are mostly related to the display of variants.

    Attributes:
        - fields (list[str]): Pre-defined fields displayed on the UI
        - source (str): Pre-defined selection of variants
        - filters (dict): Pre-defined filters
        - group_by (list[str]): Pre-defined fields used to group variants
        - having (dict): TODO
    """

    def __init__(self):

        # query
        self.fields = ["favorite", "classification", "chr", "pos", "ref", "alt", "qual"]
        self.source = "variants"
        self.filters = {}
        self.group_by = []
        self.having = {}

        # selected variants
        self.current_variant = None



"""List of classes used for settings window

A SettingsDialog is a collection of section ( SectionWidget ) 
which contains multiple page ( AbstractSettingsWidget ) to save and load settings thanks to QSettings.

* SettingsDialog: 
Main widget for settings window that instantiate all subsection widget

* SectionWidget: 
Handy class to group similar settings widgets in tabs (used by SettingsDialog).

* AbstractSettingsWidget:
Abstract class for build a page settings
    
    Subclasses:
        - TranslationSettingsWidget: Allow to choose a language for the interface
        - ProxySettingsWidget: Allow to configure proxy settings for widgets that require internet connection
        - StyleSettingsWidget
        - PluginsSettingsWidget
        - VariantSettingsWidget: Allow to add personal templates to search a variant in a third-party database

Exemples: 

    # Create sub-section  

    class MemorySettings(BaseSettings):
        def save():
            settings = self.create_settings()
            settings.setValue("value", 10)    

        def load():
            settings = self.create_settings()
            value = settings.value("value")
               

    class DiskSettings(BaseSettings):
        def save():
            settings = self.create_settings()
            settings.setValue("value", 10)    

        def load():
            settings = self.create_settings()
            value = settings.value("value")

    # create one section 
    performance_section = SectionWidget()
    performance_section.add_setting_widget(MemorySettings)
    performance_section.add_setting_widget(DiskSettings)

    # add section to the main settings widge 
    widget = SettingsWidget()
    widget.add_section(widget)

    widget.save_all()

"""
# Standard imports
import os
import glob
from abc import abstractmethod
from logging import DEBUG

# Qt imports
from PySide2.QtWidgets import *
from PySide2.QtCore import *  # QApplication.instance()
from PySide2.QtGui import *  # QIcon, QPalette

# Custom imports
import cutevariant.commons as cm
from cutevariant.gui.ficon import FIcon
from cutevariant.gui import network, style, widgets

LOGGER = cm.logger()


class AbstractSettingsWidget(QWidget):
    """Abstract class for settings widgets

    User must reimplement load() and save() 
     
    """

    def __init__(self):
        super().__init__()
        self.setWindowTitle("base")
        self.setWindowIcon(FIcon(0xF5CA))
        self.section_widget = None

    @abstractmethod
    def save(self):
        """Save the current widget settings in QSettings"""
        raise NotImplementedError(self.__class__.__name__)

    @abstractmethod
    def load(self):
        """Load settings from QSettings"""
        raise NotImplementedError(self.__class__.__name__)

    @property
    def prefix_settings(self) -> str:
        """ Return prefix settings from section parent  
        Returns:
            str
        """
        if self.section_widget:
            return self.section_widget.prefix_settings
        return None

    def create_settings(self):
        settings = QSettings()
        settings.beginGroup(self.prefix_settings)
        return settings


class SectionWidget(QTabWidget):
    """Handy class to group similar settings page in tabs"""

    def __init__(self, prefix_settings="", parent=None):
        super().__init__(parent)
        self.prefix_settings = prefix_settings

    def add_page(self, widget: AbstractSettingsWidget):
        widget.section_widget = self
        self.addTab(widget, widget.windowIcon(), widget.windowTitle())

    def save(self):
        """Call save() method of all widgets in the SectionWidget"""
        [self.widget(index).save() for index in range(self.count())]

    def load(self):
        """Call load() method of all widgets in the SectionWidget"""
        [self.widget(index).load() for index in range(self.count())]


################################################################################
class TranslationSettingsWidget(AbstractSettingsWidget):
    """Allow to choose a language for the interface"""

    def __init__(self):
        super().__init__()
        self.setWindowTitle(self.tr("Translation"))
        self.setWindowIcon(FIcon(0xF05CA))
        self.locales_combobox = QComboBox()
        mainLayout = QFormLayout()
        mainLayout.addRow(self.tr("&Choose a locale:"), self.locales_combobox)

        self.setLayout(mainLayout)
        # self.locales_combobox.currentTextChanged.connect(self.switchTranslator)

    def save(self):
        """Switch QApplication.instance() translator with the selected one and save it into config

        .. note:: settings are stored in "ui" group
        .. todo:: Handle the propagation the LanguageChange event
            https://doc.qt.io/Qt-5/qcoreapplication.html#installTranslator
            https://wiki.qt.io/How_to_create_a_multi_language_application
        """

        # Remove the old translator
        # QApplication.instance().removeTranslator(translator)

        # Load the new translator

        # Save locale setting
        locale_name = self.locales_combobox.currentText()
        settings = self.create_settings()
        settings.setValue("ui/locale", locale_name)
        app_translator = QTranslator(QApplication.instance())
        if app_translator.load(locale_name, cm.DIR_TRANSLATIONS):
            QApplication.instance().installTranslator(app_translator)

    def load(self):
        """Setup widgets in TranslationSettingsWidget"""
        self.locales_combobox.clear()
        # Get names of locales based on available files
        available_translations = {
            os.path.basename(os.path.splitext(file)[0]): file
            for file in glob.glob(cm.DIR_TRANSLATIONS + "*.qm")
        }
        # English is the default language
        available_locales = list(available_translations.keys()) + ["en"]
        self.locales_combobox.addItems(available_locales)

        # Display current locale
        settings = self.create_settings()
        locale_name = settings.value("ui/locale", "en")

        self.locales_combobox.setCurrentIndex(available_locales.index(locale_name))


class ProxySettingsWidget(AbstractSettingsWidget):
    """Allow to configure proxy settings for widgets that require internet connection"""

    def __init__(self):
        super().__init__()
        self.setWindowTitle(self.tr("Network"))
        self.setWindowIcon(FIcon(0xF0484))

        self.combo_box = QComboBox()
        self.host_edit = QLineEdit()
        self.port_edit = QSpinBox()
        self.user_edit = QLineEdit()
        self.pass_edit = QLineEdit()

        # Load proxy type
        for key in network.PROXY_TYPES:
            self.combo_box.addItem(key, network.PROXY_TYPES[key])

        # edit restriction
        self.pass_edit.setEchoMode(QLineEdit.PasswordEchoOnEdit)

        f_layout = QFormLayout()
        f_layout.addRow(self.tr("Type"), self.combo_box)
        f_layout.addRow(self.tr("Proxy host"), self.host_edit)
        f_layout.addRow(self.tr("Proxy Port"), self.port_edit)
        f_layout.addRow(self.tr("Username"), self.user_edit)
        f_layout.addRow(self.tr("Password"), self.pass_edit)

        self.combo_box.currentIndexChanged.connect(self.on_combo_changed)

        self.setLayout(f_layout)

    def save(self):
        """Save settings under "proxy" group"""
        settings = self.create_settings()
        settings.beginGroup("proxy")
        settings.setValue("type", self.combo_box.currentIndex())
        settings.setValue("host", self.host_edit.text())
        settings.setValue("port", self.port_edit.value())
        settings.setValue("username", self.user_edit.text())
        settings.setValue("password", self.user_edit.text())
        settings.endGroup()

    def load(self):
        """Load "proxy" group settings"""

        settings = self.create_settings()

        settings.beginGroup("proxy")

        s_type = settings.value("type", 0)
        if s_type:
            self.combo_box.setCurrentIndex(int(s_type))

        self.host_edit.setText(settings.value("host"))

        s_port = settings.value("port", 0)
        if s_port:
            self.port_edit.setValue(int(s_port))

        self.user_edit.setText(settings.value("username"))
        self.pass_edit.setText(settings.value("password"))
        settings.endGroup()

    def on_combo_changed(self, index):
        """ disable formular when No proxy """
        if index == 0:
            self._disable_form(True)
        else:
            self._disable_form(False)

    def _disable_form(self, disabled=True):
        """ Disabled formular """
        self.host_edit.setDisabled(disabled)
        self.port_edit.setDisabled(disabled)
        self.user_edit.setDisabled(disabled)
        self.pass_edit.setDisabled(disabled)


class StyleSettingsWidget(AbstractSettingsWidget):
    """Allow to choose a style for the interface"""

    def __init__(self):
        """Init StyleSettingsWidget

        Args:
            mainwindow (QMainWindow): Current main ui of cutevariant;
                Used to refresh the plugins
        """
        super().__init__()
        self.setWindowTitle(self.tr("Styles"))
        self.setWindowIcon(FIcon(0xF03D8))

        self.styles_combobox = QComboBox()
        mainLayout = QFormLayout()
        mainLayout.addRow(self.tr("&Choose a style:"), self.styles_combobox)

        self.setLayout(mainLayout)

    def save(self):
        """Save the selected style in config
        """
        # Get previous style

        settings = self.create_settings()

        old_style_name = settings.value("ui/style", cm.BASIC_STYLE)

        # Save style setting
        style_name = self.styles_combobox.currentText()
        if old_style_name == style_name:
            return
        settings.setValue("ui/style", style_name)

        # Change the style on the fly
        # TODO: Find a way to revert properly dark() palette and fill
        # style.bright() function.

        if style_name == cm.BASIC_STYLE:
            # Bright version: Reset style
            QApplication.instance().setStyleSheet("")
            QApplication.instance().setPalette(QApplication.style().standardPalette())
        else:
            # QApplication.instance().setStyle("fusion")
            # Apply selected style by calling on the method in style module based on its
            # name; equivalent of style.dark(app)
            getattr(style, style_name.lower())(QApplication.instance())

        # Clear pixmap cache
        QPixmapCache.clear()

        # Get the window for this widget,
        # i.e. the next ancestor widget that has a window-system frame.
        # And call refresh signal => reload the widgets
        self.window().uiSettingsChanged.emit()

    def load(self):
        """Setup widgets in StyleSettingsWidget"""
        self.styles_combobox.clear()

        # Get names of styles based on available files
        available_styles = {
            os.path.basename(os.path.splitext(file)[0]).title(): file
            for file in glob.glob(cm.DIR_STYLES + "*.qss")
            if "frameless" not in file
        }
        # Display available styles
        available_styles = list(available_styles.keys()) + [cm.BASIC_STYLE]
        self.styles_combobox.addItems(available_styles)

        # Display current style
        # Dark is the default style
        settings = self.create_settings()
        style_name = settings.value("ui/style", cm.BASIC_STYLE)
        self.styles_combobox.setCurrentIndex(available_styles.index(style_name))


class PluginsSettingsWidget(AbstractSettingsWidget):
    """Display a list of found plugin and their status (enabled/disabled)"""

    registerPlugin = Signal(dict)
    deregisterPlugin = Signal(dict)

    def __init__(self):
        super().__init__()
        self.setWindowTitle(self.tr("Plugins"))
        self.setWindowIcon(FIcon(0xF0431))
        self.view = QTreeWidget()
        self.view.setColumnCount(3)
        self.view.setHeaderLabels(["Name", "Description", "Version"])
        self.view.header().setSectionResizeMode(QHeaderView.ResizeToContents)

        main_layout = QVBoxLayout()
        main_layout.addWidget(self.view)
        self.setLayout(main_layout)

    def save(self):
        """Save the check status of enabled plugins in app settings and update UI

        Emit a `register_plugin` or `deregister_plugin` signal for the mainwindow.

        Notes:
            Called only if the user clicks on "save all" button.
        """
        for iterator in QTreeWidgetItemIterator(
            self.view, QTreeWidgetItemIterator.Enabled
        ):
            item = iterator.value()
            # Get extension and check state
            extension = item.data(0, Qt.UserRole)
            check_state = item.checkState(0) == Qt.Checked
            # Save status

            settings = self.create_settings()
            settings.setValue(f"plugins/{extension['name']}/status", check_state)

            # Set the enable status of the extension
            for sub_extension_type in {
                "widget",
                "dialog",
                "setting",
            } & extension.keys():
                extension[sub_extension_type].ENABLE = check_state

            if check_state:
                # Register plugin in UI
                self.registerPlugin.emit(extension)
            else:
                # Deregister plugin in UI
                self.deregisterPlugin.emit(extension)

    def load(self):
        """Display the plugins and their status"""
        self.view.clear()
        from cutevariant.gui import plugin

        settings = self.create_settings()

        settings_keys = set(settings.allKeys())

        for extension in plugin.find_plugins():
            displayed_title = (
                extension["name"]
                if LOGGER.getEffectiveLevel() == DEBUG
                else extension["title"]
            )
            item = QTreeWidgetItem()
            item.setText(0, displayed_title)
            item.setText(1, extension["description"])
            item.setText(2, extension["version"])

            # Is an extension enabled ?
            is_enabled = False

            # Get activation status
            # Only disabled extensions can be in settings
            key = f"plugins/{extension['name']}/status"
            activated_by_user = (
                settings.value(key) == "true" if key in settings_keys else None
            )

            for sub_extension_type in {
                "widget",
                "dialog",
                "setting",
            } & extension.keys():
                if activated_by_user is None and extension[sub_extension_type].ENABLE:
                    is_enabled = True
                    # Only disabled plugins can be reactivated by the user
                    item.setDisabled(True)
                    break
                if activated_by_user:
                    is_enabled = True
                    break

            item.setCheckState(0, Qt.Checked if is_enabled else Qt.Unchecked)
            # Attach the extension for its further activation/desactivation
            item.setData(0, Qt.UserRole, extension)

            self.view.addTopLevelItem(item)


class PathSettingsWidget(AbstractSettingsWidget):
    """ Path settings where to store shared data """

    def __init__(self):
        super().__init__()
        self.setWindowTitle(self.tr("Path"))
        self.setWindowIcon(FIcon(0xF1080))

        self.edit = widgets.FileEdit()
        self.edit.set_path_type("dir")
        main_layout = QFormLayout()
        main_layout.addRow("path", self.edit)

        self.setLayout(main_layout)

    def save(self):
        settings = QSettings()
        if self.edit.exists():
            settings.setValue("data_path", self.edit.text())

    def load(self):

        settings = QSettings()
        path = settings.value(
            "data_path",
            QStandardPaths.writableLocation(QStandardPaths.GenericDataLocation),
        )

        self.edit.setText(path)


class SettingsDialog(QDialog):
    """Main widget for settings window

    Subwidgets are intantiated on panels; a SectionWidget groups similar widgets
    in tabs.

    Signals:
        uiSettingsChanged(Signal): Emitted when some settings of the GUI are
            modified and need a reload of all widgets to take effect.
    """

    uiSettingsChanged = Signal()

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setWindowTitle(self.tr("Cutevariant - Settings"))
        self.setWindowIcon(QIcon(cm.DIR_ICONS + "app.png"))

        self.widgets = []

        self.list_widget = QListWidget()
        self.stack_widget = QStackedWidget()
        self.button_box = QDialogButtonBox(
            QDialogButtonBox.SaveAll | QDialogButtonBox.Cancel | QDialogButtonBox.Reset
        )

        self.list_widget.setFixedWidth(200)
        self.list_widget.setIconSize(QSize(32, 32))

        h_layout = QHBoxLayout()
        h_layout.addWidget(self.list_widget)
        h_layout.addWidget(self.stack_widget)

        v_layout = QVBoxLayout()
        v_layout.addLayout(h_layout)
        v_layout.addWidget(self.button_box)
        self.setLayout(v_layout)

        # Instantiate subwidgets on panels
        # Similar widgets for general configuration
        general_settings = SectionWidget()  #  Not a title !
        general_settings.setWindowTitle(self.tr("General"))
        general_settings.setWindowIcon(FIcon(0xF0614))

        general_settings.add_page(PathSettingsWidget())

        # Cutevariant is  not yet translated ..
        # general_settings.add_page(TranslationSettingsWidget())

        general_settings.add_page(ProxySettingsWidget())
        general_settings.add_page(StyleSettingsWidget())

        # Activation status of plugins
        plugin_settings = PluginsSettingsWidget()

        #  BOF...
        if parent:
            plugin_settings.registerPlugin.connect(parent.register_plugin)
            plugin_settings.deregisterPlugin.connect(parent.deregister_plugin)

        # Specialized widgets on panels
        self.add_section(general_settings)
        # self.add_section(plugin_settings)
        self.load_plugins()

        self.resize(800, 400)

        self.button_box.button(QDialogButtonBox.SaveAll).clicked.connect(self.save_all)
        self.button_box.button(QDialogButtonBox.Reset).clicked.connect(self.load_all)
        self.button_box.button(QDialogButtonBox.Cancel).clicked.connect(self.close)

        # Connection events
        self.list_widget.currentRowChanged.connect(self.stack_widget.setCurrentIndex)

        # Load settings
        self.load_all()

        self.accepted.connect(self.close)

    def add_section(self, widget: SectionWidget):
        """Add a widget on the widow via a QStackedWidget; keep a reference on it
        for later connection/activation"""
        # Used to load/save all widgets on demand
        self.widgets.append(widget)
        # Used for ui positionning and connection events
        self.list_widget.addItem(
            QListWidgetItem(widget.windowIcon(), widget.windowTitle())
        )
        self.stack_widget.addWidget(widget)

    def save_all(self):
        """Call save() method of all widgets"""
        [widget.save() for widget in self.widgets]
        self.accepted.emit()

    def load_all(self):
        """Call load() method of all widgets"""
        [widget.load() for widget in self.widgets]

    def load_plugins(self):
        """ Add plugins settings """
        from cutevariant.gui import plugin

        for extension in plugin.find_plugins():

            if "setting" in extension:
                settings_widget_class = extension["setting"]
                if not settings_widget_class.ENABLE:
                    # Skip disabled plugins
                    continue

                widget = settings_widget_class()
                # Create rprefix settings ! For instance [VariantView]
                # widget.prefix_settings = widget.__class__.__name__.replace(
                #     "SettingsWidget", ""
                # )

                if not widget.windowTitle():
                    widget.setWindowTitle(extension["name"])

                if not widget.windowIcon():
                    widget.setWindowIcon(FIcon(0xF0431))

                self.add_section(widget)


if __name__ == "__main__":
    import sys

    app = QApplication(sys.argv)

    d = SettingsDialog()
    d.show()

    app.exec_()



"""Proof of concept for the VQL DSL.

From this module, you only need to use parse_vql or parse_one_vql.
For instance::

    cmd = parse_one_vql("SELECT chr, pos FROM variants")
    print(cmd)

See test_vql.py for usage and features.

"""
import textx
from pkg_resources import resource_string


def model_class(name: str, bases: tuple, attrs: dict) -> type:
    """Metaclass to automatically build the __init__ to get the properties,
    and register the class for metamodel
    """
    if "__init__" not in attrs:

        def __init__(self, *args, **kwargs):
            for field, value in kwargs.items():
                setattr(self, field, value)

        attrs["__init__"] = __init__
    cls = type(name, bases, attrs)
    model_class.classes.append(cls)
    return cls


model_class.classes = []


class VQLSyntaxError(ValueError):
    def __init__(self, message, col=None, *args, **kwargs):
        super().__init__(message, col, *args, **kwargs)
        self.message = message
        self.col = col

    def __repr__(self):
        if self.col:
            return "VQLSyntaxError: '%s' at position %s" % (self.message, self.col)
        return "VQLSyntaxError: '%s'" % self.message

    def __str__(self):
        return self.__repr__()


# ============ Error handle ==================================
def error_message_from_err(
    err: textx.exceptions.TextXSyntaxError, raw_vql: str
) -> (str, int):
    """Return human-readable information and index in raw_sql query
    about the given exception"""
    # print(err)
    # print(dir(err))
    # print(err.args)
    # print(err.err_type)
    # print(err.line, err.col)
    # print(err.message)
    # print(err.filename)
    # print(err.expected_rules)
    if "'SELECT'" in err.message:  # was awaiting for a SELECT clause
        return "no SELECT clause", -1
    if err.message.endswith("=> 's,ref FROM*'."):
        return "empty 'FROM' clause", err.col
    if (
        ",*," in err.message
        and len(err.expected_rules) == 1
        and type(err.expected_rules[0]).__name__ == "RegExMatch"
    ):
        return "invalid empty identifier in SELECT clause", err.col
    if "Expected INT " in err.message and len(err.expected_rules) == 3:
        return "invalid value in WHERE clause", err.col
    if "Expected '==|>=|<=|!=" in err.message and len(err.expected_rules) == 1:
        return "invalid operator in WHERE clause", err.col

    raise err  # error not handled. Just raise it


# ============ Different class to map with VQL model


class FilterTerm(metaclass=model_class):
    @property
    def value(self):
        field = self.field.value if hasattr(self.field, "value") else self.field
        val = self.val.value if hasattr(self.val, "value") else self.val
        return {"field": field, "operator": self.op, "value": val}


class FilterExpression(metaclass=model_class):
    @property
    def value(self):
        out = []
        key = "AND"  # By default
        for i in self.op:
            if isinstance(i, str):
                if i in ("AND", "OR"):
                    key = i
                else:
                    out.append(i)
            else:
                out.append(i.value)
        return {key: out}


class SetExpression(metaclass=model_class):
    @property
    def value(self):
        return "test"


class FilterOperand(metaclass=model_class):
    @property
    def value(self):
        return self.op.value


class Function(metaclass=model_class):
    @property
    def value(self):
        if not self.extra:
            self.extra = "gt"
        return (self.func, self.arg, self.extra)


class Tuple(metaclass=model_class):
    @property
    def value(self):
        return tuple(self.items)


class WordSetIdentifier(metaclass=model_class):
    @property
    def value(self):
        return ("WORDSET", self.arg)


class SelectCmd(metaclass=model_class):
    @property
    def value(self):
        output = {
            "cmd": "select_cmd",
            "fields": [
                col.value if hasattr(col, "value") else col for col in self.fields
            ],
            "source": self.source,
        }

        if self.filter:
            output["filters"] = self.filter.value
        else:
            output["filters"] = {}

        if self.group_by:
            output["group_by"] = [
                i.value if isinstance(i, Function) else i for i in self.group_by
            ]

        else:
            output["group_by"] = []

        if self.having_op and self.having_val:
            output["having"] = {"op": self.having_op, "value": self.having_val}
        else:
            output["having"] = {}

        return output


class CreateCmd(metaclass=model_class):
    @property
    def value(self):
        return {
            "cmd": "create_cmd",
            "source": self.source,
            "filters": self.filter.value if self.filter else {},
            "target": self.target,
        }


class SetCmd(metaclass=model_class):
    @property
    def value(self):
        return {
            "cmd": "set_cmd",
            "target": self.target,
            "first": self.first,
            "operator": self.op,
            "second": self.second,
        }


class BedCmd(metaclass=model_class):
    @property
    def value(self):
        return {
            "cmd": "bed_cmd",
            "target": self.target,
            "source": self.source,
            "path": self.path,
        }


class CopyCmd(metaclass=model_class):
    @property
    def value(self):
        return {
            "cmd": "create_cmd",
            "source": self.source,
            "filters": {},
            "target": self.target,
        }


class CountCmd(metaclass=model_class):
    @property
    def value(self):
        obj = {
            "cmd": "count_cmd",
            "source": self.source,
            "filters": self.filters.value if self.filters else {},
        }

        return obj


class DropCmd(metaclass=model_class):
    @property
    def value(self):
        return {"cmd": "drop_cmd", "feature": self.feature, "name": self.name}


class ShowCmd(metaclass=model_class):
    @property
    def value(self):
        return {"cmd": "show_cmd", "feature": self.feature}


class ImportCmd(metaclass=model_class):
    @property
    def value(self):
        return {
            "cmd": "import_cmd",
            "feature": self.feature,
            "path": self.path,
            "name": self.name,
        }


METAMODEL = textx.metamodel_from_str(
    resource_string(__name__, "vql.tx").decode(),  # grammar extraction from vql.tx
    classes=model_class.classes,
    debug=False,
    ignore_case=True,
)


def parse_vql(raw_vql: str) -> list:
    """Execute multiline VQL statement separated by ";"

    Returns:
         (generator[dict]): yield 1 VQL object (a dictionnary) per command

    Example of VQL object::

        {
            'cmd': 'select_cmd',
            'columns': ['chr','pos'],
            'source':'variants',
            'filter': 'None'
        }
    """
    try:
        raw_model = METAMODEL.model_from_str(raw_vql)
    except textx.exceptions.TextXSyntaxError as err:
        raise VQLSyntaxError(*error_message_from_err(err, raw_vql))

    yield from (command.value for command in raw_model.commands)


def parse_one_vql(raw_vql: str) -> dict:
    """Execute 1 VQL statement

    Returns:
        (dict): 1 VQL object

    Examples of VQL object::

        {
            'cmd': 'select_cmd',
            'columns': ['chr','pos'],
            'source':'variants',
            'filter': 'None'
        }
    """
    return next(parse_vql(raw_vql))



class AbstractWriter:
    """Base class for all Writer required to export variants into a file or a database.

    Subclass it if you want a new file writer.

    Attributes:

        device: a file object typically returned by open("w")

    Example:

        >>> with open(filename,"rw") as file:
        ...    writer = MyWriter(file)
        ...    writer.save(conn)
    """

    def __init__(self, device):
        self.device = device

    def save(self, conn, *args, **kwargs) -> bool:
        """Dump data to a file"""
        raise NotImplementedError()
